/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */
#ifdef WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <errno.h>
#include "CzString.h"

#define DATA_SIZE 256

/* CZSTRINGの初期化 */
void CzString_init(CZSTRING *string)
{
	CzMemory_init(string);
}

/* CZSTRINGのメモリ開放 */
void CzString_delete(CZSTRING *string)
{
	CzMemory_delete(string);
}

#ifdef HAVE_ICONV
/* CZENCODINGの初期化 */
void CzEncoding_init(CZENCODING *enc, const char *tocode, const char *fromcode)
{
	enc->handle_iconv = iconv_open(tocode, fromcode);
}

/* CZENCODINGの開放 */
void CzEncoding_delete(CZENCODING *enc)
{
	iconv_close(enc->handle_iconv);
	enc->handle_iconv = (iconv_t)-1;
}

/* 文字列の文字コード変換 */
BOOL CzEncoding_convert(CZENCODING *enc, CZSTRING *dest, CZSTRING *src)
{
	char tmp_data[DATA_SIZE];
	char *in_buf = src->data;
	size_t in_size = src->size;
	char *out_buf = tmp_data;
	size_t out_size = DATA_SIZE;

	CzString_delete(dest);

	while(in_size > 0) {
		size_t n = iconv(enc->handle_iconv, &in_buf, &in_size,
				&out_buf, &out_size);
		if(n == (size_t)-1 && errno == EILSEQ) {
			CzString_add_datas(dest, tmp_data, DATA_SIZE - out_size);
			out_size = DATA_SIZE;
			out_buf = tmp_data;
			in_buf++;
			in_size--;
		} else if(n != (size_t)-1 && in_size == 0) {
			CzString_add_datas(dest, tmp_data, DATA_SIZE - out_size);
		} else if(n == (size_t)-1 && errno == EINVAL) {
			CzString_add_datas(dest, tmp_data, DATA_SIZE - out_size);
			in_size = 0;
		} else if(n == (size_t)-1 && errno == E2BIG) {
			CzString_add_datas(dest, tmp_data, DATA_SIZE - out_size);
			out_size = DATA_SIZE;
			out_buf = tmp_data;
		} else {
			return FALSE;
		}
	}
	out_buf = tmp_data;
	out_size = DATA_SIZE;

	if(iconv(enc->handle_iconv, NULL, NULL, &out_buf, &out_size) != (size_t)-1) {
		CzString_add_datas(dest, tmp_data, DATA_SIZE - out_size);
	}

	return TRUE;
}
#else
#ifdef WIN32
/* マルチバイト文字列からUTF16に変換 */
static BOOL CzEncoding_convert_multbyte_to_UTF16(CZENCODING *enc, CZSTRING *dest, CZSTRING *src);
/* UTF16からマルチバイト文字列に変換 */
static BOOL CzEncoding_convert_UTF16_to_multbyte(CZENCODING *enc, CZSTRING *dest, CZSTRING *src);
/* UTF16からUTF8に変換 */
static BOOL CzEncoding_convert_UTF16_to_UTF8(CZENCODING *enc, CZSTRING *dest, CZSTRING *src);
/* UTF8からUTF16に変換 */
static BOOL CzEncoding_convert_UTF8_to_UTF16(CZENCODING *enc, CZSTRING *dest, CZSTRING *src);

/* 文字列の文字コード変換 */
BOOL CzEncoding_convert(CZENCODING *enc, CZSTRING *dest, CZSTRING *src)
{
	switch(enc->mode) {
		case CONVERT_MODE_MULTBYTE_TO_UTF16:
			return CzEncoding_convert_multbyte_to_UTF16(enc, dest, src);
			break;
		case CONVERT_MODE_UTF16_TO_MULTBYTE:
			return CzEncoding_convert_UTF16_to_multbyte(enc, dest, src);
			break;
		case CONVERT_MODE_UTF16_TO_UTF8:
			return CzEncoding_convert_UTF16_to_UTF8(enc, dest, src);
			break;
		case CONVERT_MODE_UTF8_TO_UTF16:
			return CzEncoding_convert_UTF8_to_UTF16(enc, dest, src);
			break;
	}
	return FALSE;
}

/* マルチバイト文字列からUTF16に変換 */
static BOOL CzEncoding_convert_multbyte_to_UTF16(CZENCODING *enc, CZSTRING *dest, CZSTRING *src)
{
	int utfsize;
    CZCHAR *utfstr;
	BOOL ret;
	
	CzString_delete(dest);

	/* サイズを算出 */
	utfsize = MultiByteToWideChar(CP_ACP, 0, src->data, -1, NULL, 0);
	if(utfsize == 0) {
		return FALSE;
	}
	/* メモリ確保 */
	utfstr = (CZCHAR *)malloc(utfsize * sizeof(CZCHAR));
	if(!utfstr) {
		return FALSE;
	}
	/* 実際の変換 */
	utfsize = MultiByteToWideChar(CP_ACP, 0, src->data, -1, utfstr, utfsize);
	if(utfsize == 0) {
		free(utfstr);
		return FALSE;
	}
	utfstr[utfsize - 1] = 0;

	ret = CzString_add_datas(dest, (char*)utfstr, utfsize * sizeof(CZCHAR));
	free(utfstr);

	return ret;
}

/* UTF16からマルチバイト文字列に変換 */
static BOOL CzEncoding_convert_UTF16_to_multbyte(CZENCODING *enc, CZSTRING *dest, CZSTRING *src)
{
	int multsize;
	char *multstr;
	CZSTRING utf16str;
	BOOL ret;

	CzString_delete(dest);

	/* サイズを算出 */
	multsize = WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)src->data, -1, NULL, 0, NULL, NULL);
	if(multsize == 0) {
		CzString_delete(&utf16str);
		return FALSE;
	}
	/* メモリ確保 */
	multstr = (char*)malloc(multsize * sizeof(char));
	if(!multstr) {
		CzString_delete(&utf16str);
		return FALSE;
	}
	/* 実際の変換 */
	multsize = WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)src->data, -1, multstr, multsize, NULL, NULL);
	if(multsize == 0) {
		free(multstr);
		CzString_delete(&utf16str);
		return FALSE;
	}
	multstr[multsize - 1] = 0;

	ret = CzString_add_datas(dest, multstr, multsize);

	free(multstr);
	CzString_delete(&utf16str);

	return ret;
}

/* UTF32->UTF16変換 */
BOOL CzEncoding_convert_UTF32_to_UTF16(CZSTRING *dest, CZSTRING *src)
{
	CZCHAR *utf32str;
	size_t i;
	CZCHAR tmp;
	CZUTF16CHAR high, low;

	CzString_delete(dest);

	utf32str = (CZCHAR*)src->data;
	for(i = 0; i < src->size / sizeof(CZCHAR); i++) {
		if(utf32str[i] >= 0x10000) {	/* サロゲートペアの必要がある */
			tmp = utf32str[i] - 0x10000;
			high = (tmp >> 10) | 0xd80;
			low = (tmp & 0x3ff) | 0xdc00;
			if(!CzString_add_char(dest, high)) {
				return FALSE;
			}
			if(!CzString_add_char(dest, low)) {
				return FALSE;
			}
		} else {
			if(!CzString_add_char(dest, utf32str[i] & 0xffff)) {
				return FALSE;
			}
		}
	}
	return TRUE;
}

/* UTF16->UTF32変換 */
BOOL CzEncoding_convert_UTF16_to_UTF32(CZSTRING *dest, CZSTRING *src)
{
	CZUTF16CHAR high, low;
	CZUTF16CHAR *utfstr;
	CZCHAR tmp;
	size_t i;

	CzString_delete(dest);

	utfstr = (CZUTF16CHAR*)src->data;
	high = 0;
	for(i = 0; i < src->size / sizeof(CZUTF16CHAR); i++) {
		if(high != 0) {
			if(utfstr[i] < 0xdc00 || utfstr[i] >= 0xe000) { /* 下位サロゲートでないなら */
				return FALSE;
			}
			low = utfstr[i];
			/* サロゲートペアを変換 */
			tmp =  (((CZCHAR)high & 0x3ff) << 10) | (low & 0x3ff) + 0x10000;
			high = 0;
			if(!CzString_add_char(dest, tmp)) {
				return FALSE;
			}
		} else if(utfstr[i] >= 0xd800 && utfstr[i] < 0xdc00) { /* 上位サロゲートなら */
			high = utfstr[i];
		} else if(!CzString_add_char(dest, utfstr[i])) {
			return FALSE;
		}
	}
	return TRUE;
}

/* UTF16からUTF8に変換 */
static BOOL CzEncoding_convert_UTF16_to_UTF8(CZENCODING *enc, CZSTRING *dest, CZSTRING *src)
{
	int utfsize;
	char *utfstr;
	BOOL ret;

	CzString_delete(dest);

	/* サイズを算出 */
	utfsize = WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR)src->data, -1, NULL, 0, NULL, NULL);
	if(utfsize == 0) {
		return FALSE;
	}
	/* メモリ確保 */
	utfstr = (char *)malloc(utfsize * sizeof(char));
	if(!utfstr) {
		return FALSE;
	}
	/* 実際の変換 */
	utfsize = WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR)src->data, -1, utfstr, utfsize, NULL, NULL);
	if(utfsize == 0) {
		free(utfstr);
		return FALSE;
	}
	utfstr[utfsize - 1] = 0;

	ret = CzString_add_datas(dest, utfstr, utfsize);
	free(utfstr);

	return ret;
}

/* UTF8からUTF16に変換 */
static BOOL CzEncoding_convert_UTF8_to_UTF16(CZENCODING *enc, CZSTRING *dest, CZSTRING *src)
{
     int unisize;
     CZCHAR *unistr;
	 BOOL ret;

	 CzString_delete(dest);
     /* サイズを算出 */
     unisize = MultiByteToWideChar(CP_UTF8, 0, src->data, -1, NULL, 0);
     if(unisize == 0) {
          return FALSE;
     }

     /* メモリ確保 */
     unistr = (CZCHAR*)malloc(unisize * sizeof(CZCHAR));
     if(!unistr) {
          return FALSE;
     }

     /* 実際の変換 */
     unisize = MultiByteToWideChar(CP_UTF8, 0, src->data, -1, unistr, unisize);
     if(unisize == 0) {
          free(unistr);
          return FALSE;
     }
     unistr[unisize - 1] = 0;

	 ret = CzString_add_datas(dest, (char*)unistr, unisize * sizeof(CZCHAR));

     return ret;
}

/* 文字の追加 */
BOOL CzString_add_utf16char(CZSTRING *str, CZUTF16CHAR ch)
{
	return CzString_add_datas(str, (char*)&ch, sizeof(ch));
}
#endif
#endif

/* 文字の追加 */
BOOL CzString_add_char(CZSTRING *str, CZCHAR ch)
{
	return CzString_add_datas(str, (char*)&ch, sizeof(ch));
}

/* CZSTRINGのコピー */
BOOL CzString_copy(CZSTRING *dest, CZSTRING *src)
{
	CzString_delete(dest);

	return CzString_add_datas(dest, src->data, src->size);
}


