/*
 * Copyright (c) 2012-2013, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include "CzarBasic_vm.h"
#include "CzarBasic.h"
#include "CzarBasic_error.h"
#include "lex_analyzer.h"
#include "CzFStream.h"
#include "CzMemStream.h"
#include "parser.h"

#define DATA_SIZE 256

/**
static void push_val(VM_DATA *vm_data, UINT32 val);
static UINT32 pop_val(VM_DATA *vm_data);
***/
static void inc_pc(VM_DATA *vm_data);

static void op_push_int(VM_DATA *vm_data);
static void op_push_var(VM_DATA *vm_data);
static void op_pop_var(VM_DATA *vm_data);
static void op_add(VM_DATA *vm_data);
static void op_sub(VM_DATA *vm_data);
static void op_mul(VM_DATA *vm_data);
static void op_div(VM_DATA *vm_data);
static void op_cmp_eq(VM_DATA *vm_data);
static void op_cmp_ne(VM_DATA *vm_data);
static void op_cmp_gt(VM_DATA *vm_data);
static void op_cmp_ge(VM_DATA *vm_data);
static void op_cmp_lt(VM_DATA *vm_data);
static void op_cmp_le(VM_DATA *vm_data);
static void op_cond_jmp(VM_DATA *vm_data);
static void op_jmp(VM_DATA *vm_data);
static void op_jmp_num(VM_DATA *vm_data);
static void op_call_num(VM_DATA *vm_data);
static void op_call(VM_DATA *vm_data);
static void op_ret(VM_DATA *vm_data);
static void op_input(VM_DATA *vm_data);
static void op_stmt_prn_string(VM_DATA *vm_data);
static void op_stmt_prn(VM_DATA *vm_data);
static void op_stmt_clear(VM_DATA *vm_data);
static void op_ncond_jmp(VM_DATA *vm_data);
static void op_cmp_range(VM_DATA *vm_data);
static void op_repush(VM_DATA *vm_data);
static void op_pop_pc(VM_DATA *vm_data);
static void op_pop_dmy(VM_DATA *vm_data);
static void op_push_var2(VM_DATA *vm_data);
static void op_pop_var2(VM_DATA *vm_data);

/* tokenを符号無し整数に変換 */
static PARSE_RESULT conv_unsigned_int(TOKEN *token, int *retval)
{
	int pt = 0;

	*retval = 0;
	if(token->token[0] == '0' && token->token[1] != 0) {
		return PARSE_RESULT_FALSE;
	}
	while(token->token[pt] != 0) {
		if(!isdigit((int)token->token[pt])) {
			return PARSE_RESULT_FALSE;
		}
		*retval = *retval * 10 + token->token[pt] - '0';
		if(*retval & ~0x7fff) {
			return PARSE_RESULT_OVERFLOW;
		}
		pt++;
	}
	return PARSE_RESULT_TRUE;
}

/* 整数へ変換 */
static PARSE_RESULT parse_int(LEX_DATA *lex_data, int *retval)
{
	TOKEN *token;
	int val;
	PARSE_RESULT res;

	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type == TOKEN_TYPE_MINUS) { /* "-"なら */
		*retval = -1;
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			return PARSE_RESULT_FALSE;
		}
	} else if(token->token_type == TOKEN_TYPE_PLUS) { /* "+"なら */
		*retval = 1;
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		*retval = 1;
	}
	if(token->token_type != TOKEN_TYPE_INT) {
		lex_unget_token(lex_data, token);
		return PARSE_RESULT_FALSE;
	}
	res = conv_unsigned_int(token, &val);
	if(res == PARSE_RESULT_TRUE) {
		*retval = *retval * val;
	}
	
	return res;
}


/* vm実行 */
BOOL vm_run(VM_DATA *vm_data)
{
	vm_data->pc = 0;
	vm_data->err_code = ERR_NOERROR;

	while(1) {
		UINT32 code = vm_data->mem[vm_data->pc];
		inc_pc(vm_data);
		switch((enum opecode)code) {
			case OP_PUSH_INT:	/* 数値をスタックに積む */
				op_push_int(vm_data);
				break;
			case OP_PUSH_VAR:	/* 変数をスタックに積む */
				op_push_var(vm_data);
				break;
			case OP_POP_VAR:	/* 変数にスタックから値を代入 */
				op_pop_var(vm_data);
				break;
			case OP_ADD:		/* 加算 */
				op_add(vm_data);
				break;
			case OP_SUB:		/* 減算 */
				op_sub(vm_data);
				break;
			case OP_MUL:		/* 乗算 */
				op_mul(vm_data);
				break;
			case OP_DIV:		/* 除算 */
				op_div(vm_data);
				break;
			case OP_CMP_EQ:		/* 比較(=) */
				op_cmp_eq(vm_data);
				break;
			case OP_CMP_NE:		/* 比較(<>) */
				op_cmp_ne(vm_data);
				break;
			case OP_CMP_GT:		/* 比較(>) */
				op_cmp_gt(vm_data);
				break;
			case OP_CMP_GE:		/* 比較(>=) */
				op_cmp_ge(vm_data);
				break;
			case OP_CMP_LT:		/* 比較(<) */
				op_cmp_lt(vm_data);
				break;
			case OP_CMP_LE:		/* 比較(<=) */
				op_cmp_le(vm_data);
				break;
			case OP_COND_JMP:	/* 条件ジャンプ */
				op_cond_jmp(vm_data);
				break;
			case OP_JMP:		/* 無条件ジャンプ */
				op_jmp(vm_data);
				break;
			case OP_JMP_NUM:	/* 行番号ジャンプ */
				op_jmp_num(vm_data);
				break;
			case OP_CALL_NUM:		/* 行番号サブルーチンコール */
				op_call_num(vm_data);
				break;
			case OP_RET:		/* サブルーチンコールリターン */
				op_ret(vm_data);
				break;
			case OP_INPUT:		/* INPUTステートメント */
				op_input(vm_data);
				break;
			case OP_STMT_PRN_STRING:	/* 文字列出力 */
				op_stmt_prn_string(vm_data);
				break;
			case OP_STMT_PRN:		/* スタック上の値を出力 */
				op_stmt_prn(vm_data);
				break;
			case OP_STMT_CLEAR:		/* 変数のクリア */
				op_stmt_clear(vm_data);
				break;
			case OP_STMT_END:		/* 終了 */
				return TRUE;
				break;
			case OP_NCOND_JMP:	/* 偽のときジャンプ */
				op_ncond_jmp(vm_data);
				break;
			case OP_CALL:		/* 無条件サブルーチンコール */
				op_call(vm_data);
				break;
			case OP_CMP_RANGE:	/* 範囲内チェック */
				op_cmp_range(vm_data);
				break;
			case OP_REPUSH:		/* 再プッシュ */
				op_repush(vm_data);
				break;
			case OP_POP_PC:		/* スタック上のアドレスへジャンプ */
				op_pop_pc(vm_data);
				break;
			case OP_POP_DMY:	/* スタックの空読み */
				op_pop_dmy(vm_data);
				break;
			case OP_PUSH_VAR2:
				op_push_var2(vm_data);
				break;
			case OP_POP_VAR2:
				op_pop_var2(vm_data);
				break;
			default:
				vm_data->err_code = ERR_UNDEFINEDINSTRUCTION;
				break;
		}
		if(vm_data->err_code != ERR_NOERROR) {
			return FALSE;
		}
	}

	return TRUE;
}

/****
static void push_val(VM_DATA *vm_data, UINT32 val)
{
	vm_data->stack[vm_data->sp] = val;
	vm_data->sp++;
	if(vm_data->sp >= STACK_SIZE) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
	}
}

static UINT32 pop_val(VM_DATA *vm_data)
{
	if(vm_data->sp == 0) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return 0;
	}
	vm_data->sp--;
	return vm_data->stack[vm_data->sp];
}
***/

static void inc_pc(VM_DATA *vm_data)
{
	vm_data->pc++;
	if(vm_data->pc >= VMMEM_MAX) {
		vm_data->pc = 0;
	}
}

static void op_push_int(VM_DATA *vm_data)
{
	if(!CzStack_push_uint32(&vm_data->stack, vm_data->mem[vm_data->pc])) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
	inc_pc(vm_data);
}

static void op_push_var(VM_DATA *vm_data)
{
	if(vm_data->mem[vm_data->pc] >= VAR_SIZE) {
		vm_data->err_code = ERR_OUT_OF_RANGE;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, vm_data->var[vm_data->mem[vm_data->pc]])) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
	inc_pc(vm_data);
}

static void op_pop_var(VM_DATA *vm_data)
{
	UINT32 data = 0;

	if(vm_data->mem[vm_data->pc] >= VAR_SIZE) {
		vm_data->err_code = ERR_OUT_OF_RANGE;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, &data)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
		
	vm_data->var[vm_data->mem[vm_data->pc]] = data;
	inc_pc(vm_data);
}

static void op_add(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;
	INT32 val;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	val = v1 + v2;
	if(val > 32767) {
		vm_data->err_code = ERR_OVERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, val)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_sub(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;
	INT32 val;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	val = v1 - v2;
	if(val < -32768) {
		vm_data->err_code = ERR_OVERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, val)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_mul(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;
	INT32 val;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	val = v1 * v2;
	if(val > 32767) {
		vm_data->err_code = ERR_OVERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, val)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_div(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(v2 == 0) {
		vm_data->err_code = ERR_DIVISIONBYZERO;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 / v2)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cmp_eq(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 == v2 ? -1:0)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cmp_ne(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 != v2 ? -1:0)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cmp_gt(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 > v2 ? -1:0)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cmp_ge(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 >= v2 ? -1:0)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cmp_lt(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 < v2 ? -1:0)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cmp_le(VM_DATA *vm_data)
{
	INT32 v2;
	INT32 v1;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->stack, v1 <= v2 ? -1:0)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_cond_jmp(VM_DATA *vm_data)
{
	INT32 val;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&val)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(val == -1) {
		vm_data->pc = vm_data->mem[vm_data->pc];
		return;
	}
	inc_pc(vm_data);
}

static void op_jmp(VM_DATA *vm_data)
{
	vm_data->pc = vm_data->mem[vm_data->pc];
}

static void op_jmp_num(VM_DATA *vm_data)
{
	INT16 line_num;
	UINT32 addr, label_index, val;
	char str_label[STRING_SIZE];

	if(!CzStack_pop_uint32(&vm_data->stack, &val)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	line_num = val;
	sprintf(str_label, "%d", line_num);
	label_index = search_label(vm_data, str_label);
	if(label_index == (UINT32)-1) {
		vm_data->err_code = ERR_UNDEFINED_LINE_NUMBER;
		return;
	}
	addr = get_label_address(vm_data, label_index);
	if(addr == (UINT32)-1) {
		vm_data->err_code = ERR_UNDEFINED_LINE_NUMBER;
		return;
	}
	vm_data->pc = addr;
}

static void op_call_num(VM_DATA *vm_data)
{
	INT16 line_num;
	UINT32 addr, label_index, val;
	char str_label[STRING_SIZE];

	if(!CzStack_pop_uint32(&vm_data->stack, &val)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	line_num = val;
	sprintf(str_label, "%d", line_num);
	label_index = search_label(vm_data, str_label);
	if(label_index == (UINT32)-1) {
		vm_data->err_code = ERR_UNDEFINED_LINE_NUMBER;
		return;
	}
	addr = get_label_address(vm_data, label_index);
	if(addr == (UINT32)-1) {
		vm_data->err_code = ERR_UNDEFINED_LINE_NUMBER;
		return;
	}
	if(!CzStack_push_uint32(&vm_data->call_stack, vm_data->pc)) {
		vm_data->err_code = ERR_OUTOFMEMORY;
		return;
	}
	vm_data->pc = addr;
}

static void op_ret(VM_DATA *vm_data)
{
	UINT32 ret;
	if(!CzStack_pop_uint32(&vm_data->call_stack, &ret)) {
		vm_data->err_code = ERR_RETURNWITHOUTGOSUB;
		return;
	}
	vm_data->pc = ret;
}

static void op_input(VM_DATA *vm_data)
{
	int i;
	char buf[DATA_SIZE];
	CZFSTREAM stream;
	CZMEMSTREAM memstream;
	TOKEN *token;
	LEX_DATA lex_data;
	int line_size;
	size_t var_size;
	PARSE_RESULT ret;
	int retval;
	UINT32 var_index;

	i = 0;
	var_size = vm_data->mem[vm_data->pc];
	inc_pc(vm_data);

	CzFStream_init(&stream, stdin);
	while(1) {
		fprintf(stdout, "? ");
		fflush(stdout);
		/* 一行入力 */
		if((line_size = stream.super.lineinput((CZSTREAM*)&stream, buf, DATA_SIZE)) == -1) {
			vm_data->err_code = ERR_BUFFER_OVERFLOW;
			return;
		}
		CzMemStream_init(&memstream, buf, line_size);
		lex_data_init(&lex_data, (CZSTREAM*)&memstream);
		while(i != var_size) {
			token = lex_get_token_skip_space(&lex_data);
			if(token->token_type == TOKEN_TYPE_EOF
					|| token->token_type == TOKEN_TYPE_NL) {
				break;
			}
			lex_unget_token(&lex_data, token);
			/* トークンを整数値に変換 */
			ret = parse_int(&lex_data, &retval);
			if(ret == PARSE_RESULT_FALSE) { /* 整数でないなら */
				i = 0;
				fprintf(stdout, "Redo from start\n");
				break;
			} else if(ret == PARSE_RESULT_TRUE) { /* 整数なら */
				if(!CzStack_get_uint32(&vm_data->stack, -(int)var_size + i, &var_index)) {
					vm_data->err_code = ERR_VMINTERNAL;
					memstream.super.close((CZSTREAM*)&memstream);
					lex_data_delete(&lex_data);
					return;
				}
				vm_data->var[var_index] = retval;
				i++;
			} else {
				memstream.super.close((CZSTREAM*)&memstream);
				vm_data->err_code = ERR_BUFFER_OVERFLOW;
				lex_data_delete(&lex_data);
				return;
			}
			token = lex_get_token_skip_space(&lex_data);
			if(token->token_type != TOKEN_TYPE_COMMA) {
				lex_unget_token(&lex_data, token);
			}
		}
		memstream.super.close((CZSTREAM*)&memstream);
		lex_data_delete(&lex_data);
		if(i == var_size) {
			if(!memstream.super.eof((CZSTREAM*)&memstream)) { /* まだ入力項目があるなら */
				fprintf(stderr, "Extra ignored\n");
			}
			break;
		}
	}
	vm_data->stack.size -= var_size * sizeof(UINT32);
}

static void op_stmt_prn_string(VM_DATA *vm_data)
{
	UINT32 v;
	CZSTRING *string_pool, out_str;
	CZCHAR *utf;
	size_t i;
	UINT32 out_put;

	/* 出力先をスタックから取ってくる */
	if(!CzStack_pop_uint32(&vm_data->stack, &out_put)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(out_put != 1 && out_put != 2) {
		vm_data->err_code = ERR_VMINTERNAL;
		return;
	}

	v = vm_data->mem[vm_data->pc];
	inc_pc(vm_data);
	string_pool = (CZSTRING*)vm_data->string_pool.data;
#if defined(HAVE_ICONV)
	CzString_init(&out_str);
	if(!CzEncoding_convert(&vm_data->encUTFtoMultbyte, &out_str, &string_pool[v])) {
		CzString_delete(&out_str);
		return;
	}
	if(out_put == 1) {
		fprintf(stdout, "%s", out_str.data);
	} else {
		fprintf(stderr, "%s", out_str.data);
	}
	CzString_init(&out_str);
#elif defined(WIN32)
	if(out_put == 1) {
		fwprintf(stdout, L"%s", string_pool[v].data);
	} else {
		fwprintf(stderr, L"%s", string_pool[v].data);
	}
#else
	CzString_init(&out_str);

	utf = (CZCHAR*)string_pool[v].data;
	for(i = 0; i < string_pool[v].size / sizeof(CZCHAR); i++) {
		if(!CzString_add_data(&out_str, (char)utf[i])) {
			CzString_delete(&out_str);
			vm_data->err_code = ERR_OUTOFMEMORY;
			return;
		}
	}
	if(out_put == 1) {
		fprintf(stdout, "%s", out_str.data);
	} else {
		fprintf(stderr, "%s", out_str.data);
	}

	CzString_init(&out_str);
#endif

}

static void op_stmt_prn(VM_DATA *vm_data)
{
	INT32 v;
	UINT32 out_put;

	/* 出力する数値をスタックから取得する */
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	/* 出力先をスタックから取ってくる */
	if(!CzStack_pop_uint32(&vm_data->stack, &out_put)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(out_put != 1 && out_put != 2) {
		vm_data->err_code = ERR_VMINTERNAL;
		return;
	}
	if(out_put == 1) {
		fprintf(stdout, "%d", v);
	} else {
		fprintf(stderr, "%d", v);
	}
}

static void op_stmt_clear(VM_DATA *vm_data)
{
	int i;
	for(i = 0; i < VAR_SIZE; i++) {
		vm_data->var[i] = 0;
	}
}

static void op_ncond_jmp(VM_DATA *vm_data)
{
	UINT32 val;

	if(!CzStack_pop_uint32(&vm_data->stack, &val)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(val == 0) {
		vm_data->pc = vm_data->mem[vm_data->pc];
		return;
	}
	inc_pc(vm_data);
}

/* VM_DATAの初期化 */
void vm_data_init(VM_DATA *vm_data)
{
#ifdef HAVE_ICONV
#if SIZEOF_WCHAR_T == 2
#ifdef WORDS_BIGENDIAN
#define INNER_CODE	"UTF-16BE"
#else
#define INNER_CODE	"UTF-16LE"
#endif
#else
#ifdef WORDS_BIGENDIAN
#define INNER_CODE	"UTF-32BE"
#else
#define INNER_CODE	"UTF-32LE"
#endif
#endif
#endif
	char *token, *lang;
	char *mbc_code;
	vm_data->mem_last = 0;
	vm_data->line_num_table_last = 0;
	vm_data->err_code = ERR_NOERROR;
	vm_data->pc = 0;
#ifdef HAVE_ICONV
	mbc_code = "UTF-8";
	lang = getenv("LANG");
	if(lang) {
		token = strtok(lang, ".");
		if(token) {
			token = strtok(NULL, ".");
			if(token) {
				mbc_code = token;
			}
		}
	}
	CzEncoding_init(&vm_data->encMultbytetoUTF, INNER_CODE, mbc_code);
	CzEncoding_init(&vm_data->encUTFtoMultbyte, mbc_code, INNER_CODE);
#else
#ifdef WIN32
	vm_data->encMultbytetoUTF.mode = CONVERT_MODE_MULTBYTE_TO_UTF16;
	vm_data->encUTFtoMultbyte.mode = CONVERT_MODE_UTF16_TO_MULTBYTE;
#endif
#endif
	CzMemory_init(&vm_data->string_pool);
	CzMemory_init(&vm_data->label_table);
	CzStack_init(&vm_data->stack);
	CzStack_init(&vm_data->call_stack);
}

/* VM_DATAの後処理 */
void vm_data_delete(VM_DATA *vm_data)
{
	size_t i;
	CZSTRING *string_pool;
	LABEL_TABLE *label_table;

#ifdef HAVE_ICONV
	CzEncoding_delete(&vm_data->encMultbytetoUTF);
	CzEncoding_delete(&vm_data->encUTFtoMultbyte);
#endif
	string_pool = (CZSTRING*)vm_data->string_pool.data;
	for(i = 0; i < vm_data->string_pool.size / sizeof(CZSTRING); i++) {
		CzString_delete(&string_pool[i]);
	}
	CzMemory_delete(&vm_data->string_pool);

	label_table = (LABEL_TABLE*)vm_data->label_table.data;
	for(i = 0; i < vm_data->label_table.size / sizeof(LABEL_TABLE); i++) {
		Label_table_delete(&label_table[i]);
	}
	CzMemory_delete(&vm_data->label_table);
	CzStack_delete(&vm_data->stack);
	CzStack_delete(&vm_data->call_stack);
}

/* アドレスからエディタ上の行番号を求める */
int search_editor_line_num(VM_DATA *vm_data, UINT32 addr)
{
	int i;
	for(i = vm_data->line_num_table_last - 1; i >= 0; i--) {
		if(vm_data->line_num_table[i] <= addr) {
			return i + 1;
		}
	}
	return -1;
}

/* ラベルテーブルの初期化 */
void Label_table_init(LABEL_TABLE *label)
{
	CzString_init(&label->label);
	label->address = (UINT32)-1;
}

/* ラベルテーブルのメモリを解放 */
void Label_table_delete(LABEL_TABLE *label)
{
	CzString_delete(&label->label);
	Label_table_init(label);
}

#ifdef PRINT_LABEL_TABLE
/* ラベルテーブルの表示 */
void print_label_table(VM_DATA *vm_data)
{
	LABEL_TABLE *item;
	size_t i;

	printf("print_label_table\n");
	item = (LABEL_TABLE *)vm_data->label_table.data;
	for(i = 0; i < vm_data->label_table.size / sizeof(LABEL_TABLE); i++) {
		printf("%d:%s:%d\n", i, item[i].label.data, item[i].address);
	}
}
#endif

static void op_call(VM_DATA *vm_data)
{
	UINT32 addr;

	addr = vm_data->mem[vm_data->pc];
	inc_pc(vm_data);

	if(!CzStack_push_uint32(&vm_data->call_stack, vm_data->pc)) {
		vm_data->err_code = ERR_OUTOFMEMORY;
		return;
	}
	vm_data->pc = addr;
}

#ifdef PRINT_LINE_NUM_TABLE
/* 行番号テーブルの表示 */
void print_line_num_table(VM_DATA *vm_data)
{
	int i;

	printf("print_line_num_table\n");
	for(i = 0; i < vm_data->line_num_table_last; i++) {
		printf("%d:0x%x\n", i + 1, vm_data->line_num_table[i]);
	}
}
#endif

/* 範囲内チェック */
static void op_cmp_range(VM_DATA *vm_data)
{
	INT32 v, v1, v2;

	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v2)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, (UINT32*)&v1)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(v1 < v2) {
		if(v1 <= v && v2 >= v) {
			if(!CzStack_push_uint32(&vm_data->stack, (UINT32)-1)) {
				vm_data->err_code = ERR_STACK_OVERFLOW;
				return;
			}
		} else {
			if(!CzStack_push_uint32(&vm_data->stack, 0)) {
				vm_data->err_code = ERR_STACK_OVERFLOW;
				return;
			}
		}
	} else {
		if(v1 >= v && v2 <= v) {
			if(!CzStack_push_uint32(&vm_data->stack, (UINT32)-1)) {
				vm_data->err_code = ERR_STACK_OVERFLOW;
				return;
			}
		} else {
			if(!CzStack_push_uint32(&vm_data->stack, 0)) {
				vm_data->err_code = ERR_STACK_OVERFLOW;
				return;
			}
		}
	}
}

/* 再プッシュ */
static void op_repush(VM_DATA *vm_data)
{
	UINT32 tmpsp, size;
	UINT32 *p;
	INT32 offset;
	size_t i;

	offset = (INT32)vm_data->mem[vm_data->pc];
	inc_pc(vm_data);
	size = vm_data->mem[vm_data->pc];
	inc_pc(vm_data);

	tmpsp = vm_data->stack.size / sizeof(UINT32);
	p = (UINT32*)vm_data->stack.data;
	if((INT32)tmpsp + offset < 0) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	for(i = 0; i < size; i++) {
		if(!CzStack_push_uint32(&vm_data->stack, p[tmpsp + offset + i])) {
			vm_data->err_code = ERR_STACK_OVERFLOW;
			return;
		}
	}
}

static void op_pop_pc(VM_DATA *vm_data)
{
	UINT32 tmp;

	if(!CzStack_pop_uint32(&vm_data->stack, &tmp)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	vm_data->pc = tmp;
}

static void op_pop_dmy(VM_DATA *vm_data)
{
	UINT32 tmp;

	if(!CzStack_pop_uint32(&vm_data->stack, &tmp)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
}

static void op_push_var2(VM_DATA *vm_data)
{
	UINT32 var_index, tmp;

	if(!CzStack_pop_uint32(&vm_data->stack, &var_index)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(var_index >= VAR_SIZE) {
		vm_data->err_code = ERR_OUT_OF_RANGE;
		return;
	}
	tmp = vm_data->var[var_index]; 
	if(!CzStack_push_uint32(&vm_data->stack, tmp)) {
		vm_data->err_code = ERR_STACK_OVERFLOW;
		return;
	}
}

static void op_pop_var2(VM_DATA *vm_data)
{
	UINT32 var_index, tmp;

	if(!CzStack_pop_uint32(&vm_data->stack, &var_index)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}
	if(var_index >= VAR_SIZE) {
		vm_data->err_code = ERR_OUT_OF_RANGE;
		return;
	}
	if(!CzStack_pop_uint32(&vm_data->stack, &tmp)) {
		vm_data->err_code = ERR_STACK_UNDERFLOW;
		return;
	}

	vm_data->var[var_index] = tmp;
}
