/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#ifndef __CZARBASIC_VM_H__
#define __CZARBASIC_VM_H__

#include "CzarBasic.h"
#include "CzarBasic_error.h"
#include "CzType.h"
#include "CzString.h"
#include "CzStack.h"

#define VMMEM_MAX	65536	/* プログラムメモリサイズ */
#define STRING_POOL_MAX	65536	/* 文字列プールサイズ */
#define LINE_NUM_MAX	65536	/* 行番号テーブル数 */

#define VAR_SIZE	26	/* 変数の数 */
#define STACK_SIZE	256	/* スタックサイズ */

/* ラベルテーブル構造体 */
typedef struct label_table {
	CZSTRING label;		/* ラベル */
	UINT32 address;		/* アドレス */
} LABEL_TABLE;

/* 仮想マシンデータ構造体 */
typedef struct vm_data {
	/* プログラムメモリ */
	UINT32 mem[VMMEM_MAX];
	UINT32 mem_last;
	/* 文字列プール */
	CZMEMORY string_pool;
	/* 行番号テーブル */
	UINT32 line_num_table[LINE_NUM_MAX];
	/* 行番号テーブル使用数 */
	int line_num_table_last;
	/* ラベルテーブル */
	CZMEMORY label_table;
	/* スタック */
	CZSTACK stack;
	/* コールスタック */
	CZSTACK call_stack;
#if defined(HAVE_ICONV) || defined(WIN32)
	/* UTF16へ変換 */
	CZENCODING encMultbytetoUTF;
	/* UTF8へ変換 */
	CZENCODING encUTFtoMultbyte;
#endif
	/* エラー */
	ERROR_CODE err_code;
	/* 変数 */
	INT16 var[VAR_SIZE];
	/* プログラムカウンタ */
	UINT32 pc;
} VM_DATA;

/* VM_DATAの初期化 */
void vm_data_init(VM_DATA *vm_data);
/* VM_DATAの後処理 */
void vm_data_delete(VM_DATA *vm_data);

/* ラベルテーブルの初期化 */
void Label_table_init(LABEL_TABLE *label);
/* ラベルテーブルのメモリを解放 */
void Label_table_delete(LABEL_TABLE *label);

/* vm実行 */
BOOL vm_run(VM_DATA *vm_data);

#ifdef PRINT_LABEL_TABLE
/* ラベルテーブルの表示 */
void print_label_table(VM_DATA *vm_data);
#endif
#ifdef PRINT_LINE_NUM_TABLE
/* 行番号テーブルの表示 */
void print_line_num_table(VM_DATA *vm_data);
#endif
#ifdef PRINT_DISASM
/* ディスアセンブル */
void vm_disasm(VM_DATA *vm_data);
#endif


/* アドレスからエディタ上の行番号を求める */
int search_editor_line_num(VM_DATA *vm_data, UINT32 addr);

/* オペコード一覧 */
enum opecode {
	OP_NOP,
	OP_PUSH_INT,		/* 数値をスタックに積む */
	OP_PUSH_VAR,		/* 変数をスタックに積む */
	OP_POP_VAR,		/* 変数にスタックから値を代入 */
	OP_ADD,			/* 加算 */
	OP_SUB,			/* 減算 */
	OP_MUL,			/* 乗算 */
	OP_DIV,			/* 除算 */
	OP_CMP_EQ,		/* 比較(=) */
	OP_CMP_NE,		/* 比較(<>) */
	OP_CMP_GT,		/* 比較(>) */
	OP_CMP_GE,		/* 比較(>=) */
	OP_CMP_LT,		/* 比較(<) */
	OP_CMP_LE,		/* 比較(<=) */
	OP_COND_JMP,		/* 条件ジャンプ */
	OP_JMP,			/* 無条件ジャンプ */
	OP_JMP_NUM,		/* 行番号ジャンプ */
	OP_CALL_NUM,		/* 行番号サブルーチンコール */
	OP_RET,			/* サブルーチンコールリターン */
	OP_INPUT,		/* INPUTステートメント */
	OP_STMT_PRN_STRING,	/* 文字列出力 */
	OP_STMT_PRN,		/* スタック上の値を出力 */
	OP_STMT_CLEAR,		/* 変数のクリア */
	OP_STMT_END,		/* 終了 */
	OP_NCOND_JMP,		/* 条件ジャンプ(否定)*/
	OP_CALL,		/* 無条件サブルーチンコール */
	OP_CMP_RANGE,	/* 範囲内チェック */
	OP_REPUSH,		/* 再プッシュ */
	OP_POP_PC,		/* スタック上のアドレスへジャンプ */
	OP_POP_DMY,		/* スタックの空読み */
	OP_PUSH_VAR2,
	OP_POP_VAR2,
};

#endif
