/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include "CzarBasic_vm.h"
#include "CzarBasic.h"
#include "CzarBasic_error.h"
#include "lex_analyzer.h"
#include "CzFStream.h"
#include "CzMemStream.h"
#include "parser.h"

typedef void (*PRINT_FUNC)(VM_DATA *vm_data);

static void parse_nop(VM_DATA *vm_data);
static void parse_push_int(VM_DATA *vm_data);		/* 数値をスタックに積む */
static void parse_push_var(VM_DATA *vm_data);	/* 変数をスタックに積む */
static void parse_pop_var(VM_DATA *vm_data);		/* 変数にスタックから値を代入 */
static void parse_add(VM_DATA *vm_data);			/* 加算 */
static void parse_sub(VM_DATA *vm_data);			/* 減算 */
static void parse_mul(VM_DATA *vm_data);			/* 乗算 */
static void parse_div(VM_DATA *vm_data);			/* 除算 */
static void parse_cmp_eq(VM_DATA *vm_data);		/* 比較(=) */
static void parse_cmp_ne(VM_DATA *vm_data);		/* 比較(<>) */
static void parse_cmp_gt(VM_DATA *vm_data);		/* 比較(>) */
static void parse_cmp_ge(VM_DATA *vm_data);		/* 比較(>=) */
static void parse_cmp_lt(VM_DATA *vm_data);	/* 比較(<) */
static void parse_cmp_le(VM_DATA *vm_data);		/* 比較(<=) */
static void parse_cond_jmp(VM_DATA *vm_data);		/* 条件ジャンプ */
static void parse_jmp(VM_DATA *vm_data);			/* 無条件ジャンプ */
static void parse_jmp_num(VM_DATA *vm_data);		/* 行番号ジャンプ */
static void parse_call_num(VM_DATA *vm_data);		/* 行番号サブルーチンコール */
static void parse_ret(VM_DATA *vm_data);			/* サブルーチンコールリターン */
static void parse_input(VM_DATA *vm_data);	/* INPUTステートメント */
static void parse_prn_string(VM_DATA *vm_data);		/* 文字列出力 */
static void parse_prn(VM_DATA *vm_data);			/* スタック上の値を出力 */
static void parse_clear(VM_DATA *vm_data);		/* 変数のクリア */
static void parse_end(VM_DATA *vm_data);			/* 終了 */
static void parse_ncond_jmp(VM_DATA *vm_data);		/* 条件(偽)ジャンプ */
/* 無条件ブルーチンコール */
static void parse_call(VM_DATA *vm_data);
static void parse_cmp_range(VM_DATA *vm_data);
static void parse_repush(VM_DATA *vm_data);
static void parse_pop_pc(VM_DATA *vm_data);
static void parse_pop_dmy(VM_DATA *vm_data);
static void parse_push_var2(VM_DATA *vm_data);
static void parse_pop_var2(VM_DATA *vm_data);



static PRINT_FUNC st_opecode_list[] = {
	parse_nop, 
	parse_push_int,		/* 数値をスタックに積む */
	parse_push_var,	/* 変数をスタックに積む */
	parse_pop_var,		/* 変数にスタックから値を代入 */
	parse_add,			/* 加算 */
	parse_sub,			/* 減算 */
	parse_mul,			/* 乗算 */
	parse_div,			/* 除算 */
	parse_cmp_eq,		/* 比較(=) */
	parse_cmp_ne,		/* 比較(<>) */
	parse_cmp_gt,		/* 比較(>) */
	parse_cmp_ge,		/* 比較(>=) */
	parse_cmp_lt,		/* 比較(<) */
	parse_cmp_le,		/* 比較(<=) */
	parse_cond_jmp,		/* 条件ジャンプ */
	parse_jmp,			/* 無条件ジャンプ */
	parse_jmp_num,		/* 行番号ジャンプ */
	parse_call_num,		/* 行番号サブルーチンコール */
	parse_ret,			/* サブルーチンコールリターン */
	parse_input,		/* INPUTステートメント */
	parse_prn_string,		/* 文字列出力 */
	parse_prn,			/* スタック上の値を出力 */
	parse_clear,		/* 変数のクリア */
	parse_end,			/* 終了 */
	parse_ncond_jmp,		/* 条件(偽)ジャンプ */
	parse_call,		/* 無条件サブルーチンコール */
	parse_cmp_range,
	parse_repush,
	parse_pop_pc,
	parse_pop_dmy,
	parse_push_var2,
	parse_pop_var2,
};

/* ディスアセンブル */
void vm_disasm(VM_DATA *vm_data)
{
	for(vm_data->pc = 0; vm_data->pc < vm_data->mem_last; vm_data->pc++) {
		printf("%08X(%d) ", vm_data->pc, vm_data->pc);
		st_opecode_list[vm_data->mem[vm_data->pc]](vm_data);
	}
}

static void parse_nop(VM_DATA *vm_data)
{
	printf("NOP\n");
}

/* 数値をスタックに積む */
static void parse_push_int(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("PUSH_INT %d\n", (INT32)vm_data->mem[vm_data->pc]);
}
/* 変数をスタックに積む */
static void parse_push_var(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("PUSH_VAR %d\n", (INT32)vm_data->mem[vm_data->pc]);
}
/* 変数にスタックから値を代入 */
static void parse_pop_var(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("POP_VAR %d\n", (INT32)vm_data->mem[vm_data->pc]);
}
/* 加算 */
static void parse_add(VM_DATA *vm_data)
{
	printf("ADD\n");
}
/* 減算 */
static void parse_sub(VM_DATA *vm_data)
{
	printf("SUB\n");
}
/* 乗算 */
static void parse_mul(VM_DATA *vm_data)
{
	printf("MUL\n");
}
/* 除算 */
static void parse_div(VM_DATA *vm_data)
{
	printf("DIV\n");
}
/* 比較(=) */
static void parse_cmp_eq(VM_DATA *vm_data)
{
	printf("CMP_EQ\n");
}
/* 比較(<>) */
static void parse_cmp_ne(VM_DATA *vm_data)
{
	printf("CMP_NE\n");
}
/* 比較(>) */
static void parse_cmp_gt(VM_DATA *vm_data)
{
	printf("CMP_GT\n");
}
/* 比較(>=) */
static void parse_cmp_ge(VM_DATA *vm_data)
{
	printf("CMP_GE\n");
}
/* 比較(<) */
static void parse_cmp_lt(VM_DATA *vm_data)
{
	printf("CMP_LT\n");
}
/* 比較(<=) */
static void parse_cmp_le(VM_DATA *vm_data)
{
	printf("CMP_LE\n");
}
/* 条件ジャンプ */
static void parse_cond_jmp(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("COND_JMP 0x%x\n", vm_data->mem[vm_data->pc]);
}
/* 無条件ジャンプ */
static void parse_jmp(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("JMP 0x%x\n", vm_data->mem[vm_data->pc]);
}
/* 行番号ジャンプ */
static void parse_jmp_num(VM_DATA *vm_data)
{
	printf("JMP_NUM\n");
}

/* 行番号サブルーチンコール */
static void parse_call_num(VM_DATA *vm_data)
{
	printf("CALL_NUM\n");
}
/* サブルーチンコールリターン */
static void parse_ret(VM_DATA *vm_data)
{
	printf("RET\n");
}
/* INPUTステートメント */
static void parse_input(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("INPUT %d\n", vm_data->mem[vm_data->pc]);
}
/* 文字列出力 */
static void parse_prn_string(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("PRN_STRING %d\n", vm_data->mem[vm_data->pc]);
}
/* スタック上の値を出力 */
static void parse_prn(VM_DATA *vm_data)
{
	printf("PRN\n");
}
/* 変数のクリア */
static void parse_clear(VM_DATA *vm_data)
{
	printf("CLEAR\n");
}
/* 終了 */
static void parse_end(VM_DATA *vm_data)
{
	printf("END\n");
}

/* 条件(偽)ジャンプ */
static void parse_ncond_jmp(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("NCOND_JMP 0x%x\n", vm_data->mem[vm_data->pc]);
}
/* 無条件ブルーチンコール */
static void parse_call(VM_DATA *vm_data)
{
	vm_data->pc++;
	printf("CALL 0x%x\n", vm_data->mem[vm_data->pc]);
}

static void parse_cmp_range(VM_DATA *vm_data)
{
	printf("CMP_RANGE\n");
}

static void parse_repush(VM_DATA *vm_data)
{
	int offset;
	UINT32 size;
	vm_data->pc++;
	offset = (int)vm_data->mem[vm_data->pc];
	vm_data->pc++;
	size = vm_data->mem[vm_data->pc];
	printf("REPUSH %d, %u\n", offset, size);
}

static void parse_pop_pc(VM_DATA *vm_data)
{
	printf("POP_PC\n");
}

static void parse_pop_dmy(VM_DATA *vm_data)
{
	printf("POP_DMY\n");
}

static void parse_push_var2(VM_DATA *vm_data)
{
	printf("PUSH_VAR2\n");
}

static void parse_pop_var2(VM_DATA *vm_data)
{
	printf("POP_VAR2\n");
}

