/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lex_analyzer.h"
#include "CzFStream.h"
#include "CzarBasic.h"
#include "CzarBasic_error.h"

/* 演算子検索 */
static TOKEN_TYPE search_operator1(int ch);
static TOKEN_TYPE search_operator2(int ch1, int ch2);
/* 予約語検索 */
static TOKEN_TYPE search_keyword(char *token);


/* 記号チェック */
static int isMark(int ch);

enum token_status {
	STATUS_NONE,		/* 無し */
	STATUS_STRING,		/* 文字列 */
	STATUS_WORD,		/* 単語 */
	STATUS_INT,		/* 数値 */
};

/* 記号一覧 */
static const char *st_mark = " \t!\"#$%&'()*+,-./:;<=>?[\\]^_`{|}~";

/* キーワード */
static struct keyword {
	TOKEN_TYPE token_type;
	char *keyword;
} st_keyword[] = {
	{ TOKEN_TYPE_PRINT, "PRINT" },	/* PRINT */
	{ TOKEN_TYPE_IF, "IF" },	/* IF */
	{ TOKEN_TYPE_GOTO, "GOTO" },	/* GOTO */
	{ TOKEN_TYPE_LET, "LET" },		/* LET */
	{ TOKEN_TYPE_INPUT, "INPUT" },	/* INPUT */
	{ TOKEN_TYPE_GOSUB, "GOSUB" },	/* GOSUB */
	{ TOKEN_TYPE_RETURN, "RETURN" },	/* RETURN */
	{ TOKEN_TYPE_CLEAR, "CLEAR" },	/* CLEAR */
	{ TOKEN_TYPE_END, "END" },		/* END */
	{ TOKEN_TYPE_THEN, "THEN" },	/* THEN */
	{ TOKEN_TYPE_REM, "REM" },	/* REM */
	{ TOKEN_TYPE_REM, "'" },	/* REM */
	{ TOKEN_TYPE_ELSE, "ELSE" },	/* ELSE */
	{ TOKEN_TYPE_FOR, "FOR" },	/* FOR */
	{ TOKEN_TYPE_TO, "TO" },	/* TO */
	{ TOKEN_TYPE_STEP, "STEP" },	/* STEP */
	{ TOKEN_TYPE_NEXT, "NEXT" },	/* NEXT */
	{ TOKEN_TYPE_NONE, NULL },
};

/* １文字の演算子および1文字記号のステートメント */
static struct operator_1 {
	TOKEN_TYPE token_type;
	char *keyword;
} st_operator_1[] = {
	{ TOKEN_TYPE_SPACE, " " },
	{ TOKEN_TYPE_SPACE, "\t" },
	{ TOKEN_TYPE_LPARENTHESES, "(" },
	{ TOKEN_TYPE_RPARENTHESES, ")" },
	{ TOKEN_TYPE_PLUS, "+" },
	{ TOKEN_TYPE_MINUS, "-" },
	{ TOKEN_TYPE_MUL, "*" },
	{ TOKEN_TYPE_DIV, "/" },
	{ TOKEN_TYPE_EQ, "=" },
	{ TOKEN_TYPE_LT, "<" },
	{ TOKEN_TYPE_GT, ">" },
	{ TOKEN_TYPE_REM, "'" },
	{ TOKEN_TYPE_COLON, ":" },
	{ TOKEN_TYPE_COMMA, "," },
	{ TOKEN_TYPE_SEMICOLON, ";" },
	{ TOKEN_TYPE_SHARP, "#" },
	{ TOKEN_TYPE_NONE, NULL },
};

/* ２文字の演算子 */
static struct operator_2 {
	TOKEN_TYPE token_type;
	char *keyword;
} st_operator_2[] = {
	{ TOKEN_TYPE_NE, "<>" },
	{ TOKEN_TYPE_LE, "<=" },
	{ TOKEN_TYPE_GE, ">=" },
	{ TOKEN_TYPE_NONE, NULL},
};

/* Tokenを取得する */
TOKEN *lex_get_token(LEX_DATA *stream)
{
	int ch;
	int pt;
	int ch2;
	enum token_status status;
	TOKEN_TYPE type;

	if(stream->unget_token) {
		stream->unget_token = 0;
		return &stream->token;
	}

	pt = 0;
	status = STATUS_NONE;
	stream->token.token[0] = 0;
	while((ch = stream->stream->getch(stream->stream)) != EOF) {
		if(pt >= STRING_SIZE - 1) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			stream->token.token_type = TOKEN_TYPE_ERROR;
			return &stream->token;
		}
		switch(status) {
		case STATUS_NONE:	/* 無し */
			if(ch == '\r') { /* CRなら */
				ch2 = stream->stream->getch(stream->stream);
				if(ch2 != '\n') { /* LF以外なら */
					/* 戻す */
					stream->stream->ungetc(stream->stream, ch2);
				}
				stream->token.token_type = TOKEN_TYPE_NL;
				stream->token.token[0] = 0;
				return &stream->token;
			} else if(ch == '\n') { /* LFなら */
				stream->token.token_type = TOKEN_TYPE_NL;
				stream->token.token[0] = 0;
				return &stream->token;
			} else if(ch == '"') { /* 文字列開始なら */
				status = STATUS_STRING;
				stream->token.token_type = TOKEN_TYPE_STRING;
			} else if(isdigit(ch)) { /* 数値なら */
				status = STATUS_INT;
				stream->token.token_type = TOKEN_TYPE_INT;
				stream->token.token[pt] = (char)ch;
				pt++;
			} else if(!isMark(ch)) { /* 記号以外なら */
				status = STATUS_WORD;
				stream->token.token_type = TOKEN_TYPE_WORD;
				stream->token.token[pt] = (char)ch;
				pt++;
			} else {
				TOKEN_TYPE ope;
				stream->token.token_type = TOKEN_TYPE_MARK;
				stream->token.token[pt] = (char)ch;
				pt++;
				ch2 = stream->stream->getch(stream->stream);
				if((ope = search_operator2(ch, ch2)) != TOKEN_TYPE_NONE) {
					stream->token.token_type = ope;
					stream->token.token[pt] = (char)ch2;
					pt++;
				} else {
					stream->stream->ungetc(stream->stream, ch2);
					if((ope = search_operator1(ch)) != TOKEN_TYPE_NONE) {
						stream->token.token_type = ope;
					}
				}
				stream->token.token[pt] = 0;
				return &stream->token;
			}
			break;
		case STATUS_WORD:		/* 単語 */
			if(ch == '.' && stream->token.token[pt - 1] == '.') {
				/* 戻す */
				stream->stream->ungetc(stream->stream, ch);
				stream->token.token_type = TOKEN_TYPE_WORD;
				stream->token.token[pt] = 0;
				type = search_keyword(stream->token.token);
				if(type != TOKEN_TYPE_NONE) {
					stream->token.token_type = type;
				}
				return &stream->token;
			}

			if(isdigit(ch) || isalpha(ch) || ch == '.') { /* 数値またはアルファベットなら */
				stream->token.token[pt] = (char)ch;
				pt++;
			} else {
				/* 戻す */
				stream->stream->ungetc(stream->stream, ch);
				stream->token.token_type = TOKEN_TYPE_WORD;
				stream->token.token[pt] = 0;
				type = search_keyword(stream->token.token);
				if(type != TOKEN_TYPE_NONE) {
					stream->token.token_type = type;
				}
				return &stream->token;
			}
			break;
		case STATUS_INT:		/* 数値 */
			if(isdigit(ch)) {
				stream->token.token[pt] = (char)ch;
				pt++;
			} else {
				stream->stream->ungetc(stream->stream, ch);
				stream->token.token[pt] = 0;
				return &stream->token;
			}
			break;
		case STATUS_STRING:		/* 文字列 */
			if(ch != '"') {
				stream->token.token[pt] = (char)ch;
				pt++;
			} else {
				stream->token.token[pt] = 0;
				return &stream->token;
			}
			break;
		}
	}

	if(pt == 0) {
		stream->token.token_type = TOKEN_TYPE_EOF;
		stream->token.token[0] = 0;
		return &stream->token;
	}
	stream->token.token[pt] = 0;

	return &stream->token;
}

/* tokenを戻す */
void lex_unget_token(LEX_DATA *stream, TOKEN *token)
{
	stream->unget_token = 1;
	if(token != &stream->token) {
		memcpy(&stream->token, token, sizeof(TOKEN));
	}
}

/* １文字演算子を検索 */
static TOKEN_TYPE search_operator1(int ch)
{
	int i = 0;

	while(st_operator_1[i].token_type != TOKEN_TYPE_NONE) {
		if(st_operator_1[i].keyword[0] == (char)ch) {
			return st_operator_1[i].token_type;
		}
		i++;
	}
	return TOKEN_TYPE_NONE;
}

/* ２文字演算子を検索 */
static TOKEN_TYPE search_operator2(int ch1, int ch2)
{
	int i = 0;

	while(st_operator_2[i].token_type != TOKEN_TYPE_NONE) {
		if(st_operator_2[i].keyword[0] == (char)ch1 && 
				st_operator_2[i].keyword[1] == (char)ch2) {
			return st_operator_2[i].token_type;
		}
		i++;
	}
	return TOKEN_TYPE_NONE;
}

/* 記号チェック */
static int isMark(int ch)
{
	UINT64 i;

	for(i = 0; i < strlen(st_mark); i++) {
		if(st_mark[i] == (char)ch) {
			return 1;
		}
	}
	return 0;
}

/* LEX_DATAの初期化 */
void lex_data_init(LEX_DATA *lex_data, CZSTREAM *stream)
{
	lex_data->stream = stream;
	lex_data->err_code = ERR_NOERROR;
	lex_data->err_line_num = -1;
	CzMemory_init(&lex_data->label_address);
	lex_data->unget_token = 0;
}

/* LEX_DATAのメモリの開放 */
void lex_data_delete(LEX_DATA *lex_data)
{
	CzMemory_init(&lex_data->label_address);
}

/* 予約語検索 */
static TOKEN_TYPE search_keyword(char *token)
{
	int i;

	i = 0;
	while(st_keyword[i].token_type != TOKEN_TYPE_NONE) {
		if(Strcmp(st_keyword[i].keyword, token) == 0) {
			return st_keyword[i].token_type;
		}
		i++;
	}
	return TOKEN_TYPE_NONE;
}

/* 文字列比較 */
int Strcmp(const char *str1, const char *str2)
{
	UINT64 i;

	if(strlen(str1) != strlen(str2)) {
		return -1;
	}

	for(i = 0; i < strlen(str1); i++) {
		int ch1 = toupper((int)str1[i]);
		int ch2 = toupper((int)str2[i]);
		if(ch1 != ch2) {
			return -1;
		}
	}
	return 0;
}

/* Tokenを取得する(スペースは飛ばす) */
TOKEN *lex_get_token_skip_space(LEX_DATA *stream)
{
	TOKEN *token;

	do {
		token = lex_get_token(stream);
	} while(token->token_type == TOKEN_TYPE_SPACE);
//	fprintf(stderr, "gettoken:");
//	dprint_token(token);

	return token;
}

#ifdef LEX_TEST
int lex_test(int argc, char *argv[])
{
	TOKEN *token;
	CZFSTREAM stream;
	LEX_DATA lex_data;
	FILE *fp;

	fprintf(stderr, "LEX_TEST\n");
	fp = fopen(argv[1], "r");
	CzFStream_init(&stream, fp);
	lex_data_init(&lex_data, (CZSTREAM*)&stream);
	while((token = lex_get_token(&lex_data)) != NULL) {
		if(token->token_type == TOKEN_TYPE_EOF) {
			break;
		}
		printf("-----\n");
		printf("token_type:");
		switch(token->token_type) {
		case TOKEN_TYPE_NL:		/* 改行 */
			printf("改行\n");
			break;
		case TOKEN_TYPE_WORD:		/* 単語 */
			printf("単語\n");
			break;
		case TOKEN_TYPE_SPACE:	/* スペース */
			printf("スペース\n");
			break;
		case TOKEN_TYPE_MARK:	/* 記号 */
			printf("記号\n");
			break;
		case TOKEN_TYPE_STRING:	/* 文字列 */
			printf("文字列\n");
			break;
		case TOKEN_TYPE_LPARENTHESES:	/* ( */
			printf("(\n");
			break;
		case TOKEN_TYPE_RPARENTHESES:	/* ) */
			printf(")\n");
			break;
		case TOKEN_TYPE_PLUS:		/* + */
			printf("+\n");
			break;
		case TOKEN_TYPE_MINUS:		/* - */
			printf("-\n");
			break;
		case TOKEN_TYPE_MUL:		/* * */
			printf("*\n");
			break;
		case TOKEN_TYPE_DIV:		/* / */
			printf("/\n");
			break;
		case TOKEN_TYPE_EQ:		/* = */
			printf("=\n");
			break;
		case TOKEN_TYPE_GT:		/* > */
			printf(">\n");
			break;
		case TOKEN_TYPE_GE:		/* >= */
			printf(">=\n");
			break;
		case TOKEN_TYPE_LT:		/* < */
			printf("<\n");
			break;
		case TOKEN_TYPE_LE:		/* <= */
			printf("<=\n");
			break;
		case TOKEN_TYPE_NE:		/* <> */
			printf("<>\n");
			break;
		case TOKEN_TYPE_PRINT:	/* PRINT */
			printf("PRINT\n");
			break;
		case TOKEN_TYPE_IF:		/* IF */
			printf("PRINT\n");
			break;
		case TOKEN_TYPE_GOTO:	/* GOTO */
			printf("PRINT\n");
			break;
		case TOKEN_TYPE_LET:		/* LET */
			printf("LET\n");
			break;
		case TOKEN_TYPE_INPUT:	/* INPUT */
			printf("INPUT\n");
			break;
		case TOKEN_TYPE_GOSUB:	/* GOSUB */
			printf("GOSUB\n");
			break;
		case TOKEN_TYPE_RETURN:	/* RETURN */
			printf("RETURN\n");
			break;
		case TOKEN_TYPE_CLEAR:	/* CLEAR */
			printf("CLEAR\n");
			break;
		case TOKEN_TYPE_END:		/* END */
			printf("END\n");
			break;
		case TOKEN_TYPE_THEN:	/* THEN */
			printf("THEN\n");
			break;
		case TOKEN_TYPE_REM:	/* REM */
			printf("REM\n");
			break;
		case TOKEN_TYPE_ELSE:	/* ELSE */
			printf("ELSE\n");
			break;
		case TOKEN_TYPE_FOR:	/* FOR */
			printf("FOR\n");
			break;
		case TOKEN_TYPE_TO:	/* TO */
			printf("TO\n");
			break;
		case TOKEN_TYPE_STEP:	/* STEP */
			printf("STEP\n");
			break;
		case TOKEN_TYPE_INT:	/* INT */
			printf("数値\n");
			break;
		default:
			printf("ERROR\n");
			break;
		}
		printf("token:'%s'\n", token->token);
	}
	stream.super.close((CZSTREAM*)&stream);
	lex_data_delete(&lex_data);
	fprintf(stderr, "LEX_TEST END\n");
	return 0;
}
#endif

void dprint_token(TOKEN *token)
{
	switch(token->token_type) {
	case TOKEN_TYPE_NONE:
		fprintf(stderr, "TOKEN_TYPE_NONE:");
		break;
	case TOKEN_TYPE_ERROR:
		fprintf(stderr, "TOKEN_TYPE_ERROR:");
		break;
	case TOKEN_TYPE_EOF:
		fprintf(stderr, "TOKEN_TYPE_EOF:");
		break;
	case TOKEN_TYPE_NL:		/* 改行 */
		fprintf(stderr, "TOKEN_TYPE_NL:");
		break;
	case TOKEN_TYPE_WORD:	/* 単語 */
		fprintf(stderr, "TOKEN_TYPE_WORD:");
		break;
	case TOKEN_TYPE_INT:		/* 数値 */
		fprintf(stderr, "TOKEN_TYPE_INT:");
		break;
	case TOKEN_TYPE_SPACE:	/* スペース */
		fprintf(stderr, "TOKEN_TYPE_SPACE:");
		break;
	case TOKEN_TYPE_MARK:	/* 記号 */
		fprintf(stderr, "TOKEN_TYPE_MARK:");
		break;
	case TOKEN_TYPE_STRING:	/* 文字列 */
		fprintf(stderr, "TOKEN_TYPE_STRING:");
		break;
	case TOKEN_TYPE_LPARENTHESES:	/* ( */
		fprintf(stderr, "TOKEN_TYPE_LPARENTHESES:");
		break;
	case TOKEN_TYPE_RPARENTHESES:	/* ) */
		fprintf(stderr, "TOKEN_TYPE_RPARENTHESES:");
		break;
	case TOKEN_TYPE_PLUS:		/* + */
		fprintf(stderr, "TOKEN_TYPE_PLUS:");
		break;
	case TOKEN_TYPE_MINUS:		/* - */
		fprintf(stderr, "TOKEN_TYPE_MINUS:");
		break;
	case TOKEN_TYPE_MUL:		/* * */
		fprintf(stderr, "TOKEN_TYPE_MUL:");
		break;
	case TOKEN_TYPE_DIV:		/* / */
		fprintf(stderr, "TOKEN_TYPE_DIV:");
		break;
	case TOKEN_TYPE_EQ:		/* = */
		fprintf(stderr, "TOKEN_TYPE_EQ:");
		break;
	case TOKEN_TYPE_GT:		/* > */
		fprintf(stderr, "TOKEN_TYPE_GT:");
		break;
	case TOKEN_TYPE_GE:		/* >= */
		fprintf(stderr, "TOKEN_TYPE_GE:");
		break;
	case TOKEN_TYPE_LT:		/* < */
		fprintf(stderr, "TOKEN_TYPE_LT:");
		break;
	case TOKEN_TYPE_LE:		/* <= */
		fprintf(stderr, "TOKEN_TYPE_LE:");
		break;
	case TOKEN_TYPE_NE:		/* <> */
		fprintf(stderr, "TOKEN_TYPE_NE:");
		break;
	case TOKEN_TYPE_COLON:		/* : */
		fprintf(stderr, "TOKEN_TYPE_COLON:");
		break;
	case TOKEN_TYPE_COMMA:		/* , */
		fprintf(stderr, "TOKEN_TYPE_COMMA:");
		break;
	case TOKEN_TYPE_SEMICOLON:		/* ; */
		fprintf(stderr, "TOKEN_TYPE_SEMICOLON:");
		break;
	case TOKEN_TYPE_SHARP:	/* # */
		fprintf(stderr, "TOKEN_TYPE_SHARP:");
		break;

	case TOKEN_TYPE_PRINT:	/* PRINT */
		fprintf(stderr, "TOKEN_TYPE_PRINT:");
		break;
	case TOKEN_TYPE_IF:		/* IF */
		fprintf(stderr, "TOKEN_TYPE_IF:");
		break;
	case TOKEN_TYPE_GOTO:	/* GOTO */
		fprintf(stderr, "TOKEN_TYPE_GOTO:");
		break;
	case TOKEN_TYPE_LET:		/* LET */
		fprintf(stderr, "TOKEN_TYPE_LET:");
		break;
	case TOKEN_TYPE_INPUT:	/* INPUT */
		fprintf(stderr, "TOKEN_TYPE_INPUT:");
		break;
	case TOKEN_TYPE_GOSUB:	/* GOSUB */
		fprintf(stderr, "TOKEN_TYPE_GOSUB:");
		break;
	case TOKEN_TYPE_RETURN:	/* RETURN */
		fprintf(stderr, "TOKEN_TYPE_RETURN:");
		break;
	case TOKEN_TYPE_CLEAR:	/* CLEAR */
		fprintf(stderr, "TOKEN_TYPE_CLEAR:");
		break;
	case TOKEN_TYPE_END:		/* END */
		fprintf(stderr, "TOKEN_TYPE_END:");
		break;
	case TOKEN_TYPE_THEN:	/* THEN */
		fprintf(stderr, "TOKEN_TYPE_THEN:");
		break;
	case TOKEN_TYPE_REM:		/* REM */
		fprintf(stderr, "TOKEN_TYPE_REM:");
		break;
	case TOKEN_TYPE_ELSE:	/* ELSE */
		fprintf(stderr, "TOKEN_TYPE_ELSE:");
		break;
	case TOKEN_TYPE_FOR:		/* FOR */
		fprintf(stderr, "TOKEN_TYPE_FOR:");
		break;
	case TOKEN_TYPE_TO:		/* TO */
		fprintf(stderr, "TOKEN_TYPE_TO:");
		break;
	case TOKEN_TYPE_STEP:	/* STEP */
		fprintf(stderr, "TOKEN_TYPE_STEP:");
		break;
	case TOKEN_TYPE_NEXT:	/* NEXT */
		fprintf(stderr, "TOKEN_TYPE_NEXT:");
		break;
	default:
		fprintf(stderr, "TOKEN:%d:", token->token_type);
	}
	fprintf(stderr, "%s\n", token->token);
}
