/*
 * Copyright (c) 2012-2013, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "parser.h"
#include "lex_analyzer.h"
#include "CzarBasic_vm.h"
#include "CzarBasic_error.h"

/*
#define DPRINT(a)	fprintf(stderr, "%d:%s\n", lex_data->err_line_num, a)
*/
#define DPRINT(a)

/* 文字列プールポインタ */
static int st_string_pt;
static UINT32 NL_STRING_PT;
static UINT32 TAB_STRING_PT;
static UINT32 NEXT_ERROR_STRING_PT;

/* <プログラム> ->  ε | <行> <プログラム> */
static PARSE_RESULT parser_program(VM_DATA *vm_data, LEX_DATA *lex_data);
/*
<行> -> <行番号> <行1> <行2> <CR>
<行> -> ‘#’ <全文字>* <CR>          <- <CR>までスキップする処理をここで行う。
*/
static PARSE_RESULT parser_line(VM_DATA *vm_data, LEX_DATA *lex_data);

/* <行番号> -> ε | <符号無し整数> */
static PARSE_RESULT parser_linenum(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <行1> -> ε | <ラベル> ‘:’ */
static PARSE_RESULT parser_line1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <行2> -> ε | <命令リスト> */
static PARSE_RESULT parser_line2(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <行3> -> ε | ‘:’ */
static PARSE_RESULT parser_line3(VM_DATA *vm_data, LEX_DATA *lex_data);

/* <ラベル> -> ‘*’ <WORD> */
static PARSE_RESULT parser_label(VM_DATA *vm_data, LEX_DATA *lex_data, int *ret_index);
/* <命令リスト> -> <命令リスト2> <命令リスト1> */
static PARSE_RESULT parser_statement_list(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <項1_1> -> “+” | “-" */
static PARSE_RESULT parser_item1_1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <IF文1> -> ε | ‘ELSE’ <IF文2> */
static PARSE_RESULT parser_if1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <IF文2> -> <符号無し整数> | <命令リスト> */
static PARSE_RESULT parser_if2(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <IF文3> -> ‘THEN' <IF文2> | ‘GOTO’ <飛び先> */
static PARSE_RESULT parser_if3(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <式リスト> -> ε | <式リスト1> ?<式リスト4> */
static PARSE_RESULT parser_expression_list(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl);
static PARSE_RESULT parser_expression_list1(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl);
/* <式リスト2> -> “,” | “;” */
static PARSE_RESULT parser_expression_list2(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl);
/* <式リスト3> -> ε  | <式リスト1> */
static PARSE_RESULT parser_expression_list3(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl);
/* <式リスト4> -> ε  | <式リスト2> <式リスト3> <式リスト4> */
static PARSE_RESULT parser_expression_list4(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl);

/*
<命令> -> <PRINT文> |
                  <IF文> |
                  <GOTO文> |
                  <LET文> |
                  <INPUT文> |
                  <GOSUB文> |
                  <RETURN文> |
                  <CLEAR文> |
                  <END文> |
                  <代入文> |
                  <REM文> |
		  <FOR文> |
		  <NEXT文>
*/
static PARSE_RESULT parser_statement(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <命令リスト1> -> ε |  ‘:’ <命令リスト2> <命令リスト1> */
static PARSE_RESULT parser_statement_list1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <命令リスト2> -> ε | <命令> */
static PARSE_RESULT parser_statement_list2(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <代入文> -> <変数> ‘=' <式> */
static PARSE_RESULT parser_statement_assignment(VM_DATA *vm_data, LEX_DATA *lex_data, int *var_index);
/* <PRINT文> -> ‘PRINT' <式リスト> */
static PARSE_RESULT parser_print(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <IF文> -> ‘IF' <式> <IF文3> <IF文1> */
static PARSE_RESULT parser_if(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <GOTO文> -> 'GOTO’ <飛び先> */
static PARSE_RESULT parser_goto(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <LET文> -> 'LET' <変数> ‘=' <式> */
static PARSE_RESULT parser_let(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <INPUT文> -> 'INPUT' <変数リスト> */
static PARSE_RESULT parser_input(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <GOSUB文> -> ‘GOSUB' <飛び先> */
static PARSE_RESULT parser_gosub(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <RETURN文> -> 'RETURN' */
static PARSE_RESULT parser_return(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <CLEAR文> -> ‘CLEAR' */
static PARSE_RESULT parser_clear(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <END文> -> 'END' */
static PARSE_RESULT parser_end(VM_DATA *vm_data, LEX_DATA *lex_data);
static PARSE_RESULT parser_rem(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <FOR文> -> ‘FOR’ <代入文> ‘TO’ <式> <FOR文1> */
static PARSE_RESULT parser_for(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <FOR文1> -> ε | ‘STEP’ <式> */
static PARSE_RESULT parser_for1(VM_DATA *vm_data, LEX_DATA *lex_data, int var_index);
/* <NEXT文> -> ‘NEXT’ <NEXT文1> */
static PARSE_RESULT parser_next(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <NEXT文1> -> ε | <変数リスト> */
/*
static PARSE_RESULT parser_next1(VM_DATA *vm_data, LEX_DATA *lex_data);
*/
/* <変数> -> <WORD>                                           <-- <WORD>は1文字かつアルファベットである必要がある。 */
static PARSE_RESULT parser_var(VM_DATA *vm_data, LEX_DATA *lex_data, int *var_index);
/* <式> -> <項1>  <式2> */
static PARSE_RESULT parser_expression(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <項1> -> <項2> <項1_2> */
static PARSE_RESULT parser_item1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <式2> -> ε | <式1> <項1> <式2> */
static PARSE_RESULT parser_expression2(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <項2> -> <基本式> <項2_2> */
static PARSE_RESULT parser_item2(VM_DATA *vm_data, LEX_DATA *lex_data);
static PARSE_RESULT parser_item1_2(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <基本式> -> <整数> | "(" <式> ")" | <変数> */
static PARSE_RESULT parser_basis_expression(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <項2_2> -> ε | <項2_1> <基本式> <項2_2> */
static PARSE_RESULT parser_item2_2(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <整数> -> <整数1> <符号無し整数> */
static PARSE_RESULT parser_int(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <整数1> -> ε | <符号> */
/*
static PARSE_RESULT parser_int1(VM_DATA *vm_data, LEX_DATA *lex_data);
*/
/*
static PARSE_RESULT parser_code(VM_DATA *vm_data, LEX_DATA *lex_data);
*/
static PARSE_RESULT parser_item2_1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <式1> -> “=“ | ”<“ | ”>” | "<=“ | ">=“ | "<>” */
static PARSE_RESULT parser_expression1(VM_DATA *vm_data, LEX_DATA *lex_data);
/* <変数リスト> ->  <変数> <変数リスト1> */
static PARSE_RESULT parser_var_list(VM_DATA *vm_data, LEX_DATA *lex_data, int *ret_count);
static PARSE_RESULT parser_var_list1(VM_DATA *vm_data, LEX_DATA *lex_data, int *ret_count);
/* <飛び先> -> <式> | <ラベル> */
static PARSE_RESULT parser_jump(VM_DATA *vm_data, LEX_DATA *lex_data, int *label_index);

typedef PARSE_RESULT (*STATEMENTFUNC)(VM_DATA *vm_data, LEX_DATA *);
static STATEMENTFUNC st_statement_func[] = {
	parser_print,
	parser_if,
	parser_goto,
	parser_let,
	parser_input,
	parser_gosub,
	parser_return,
	parser_clear,
	parser_end,
	NULL,		/* THEN */
	parser_rem,
	NULL,		/* ELSE */
	parser_for,
	NULL,		/* TO */
	NULL,		/* STEP */
	parser_next,
};

static struct __mark_item {
	TOKEN_TYPE mark;	/* トークンタイプ */
	enum opecode opecode;	/* オペコード */
} st_mark_list[] = {
	{ TOKEN_TYPE_PLUS, OP_ADD },		/* + */
	{ TOKEN_TYPE_MINUS, OP_SUB },		/* - */
	{ TOKEN_TYPE_MUL, OP_MUL },		/* * */
	{ TOKEN_TYPE_DIV, OP_DIV },		/* / */
	{ TOKEN_TYPE_EQ, OP_CMP_EQ },		/* = */
	{ TOKEN_TYPE_GT, OP_CMP_GT },		/* > */
	{ TOKEN_TYPE_GE, OP_CMP_GE },		/* >= */
	{ TOKEN_TYPE_LT, OP_CMP_LT },		/* < */
	{ TOKEN_TYPE_LE, OP_CMP_LE },		/* <= */
	{ TOKEN_TYPE_NE, OP_CMP_NE },		/* <> */
	{ TOKEN_TYPE_NONE, OP_NOP },
};


/* 改行チェック */
/* 戻り値:-1 エラー */
/*         0 改行以外 */
/*         1 改行 */
static int check_token_NL(TOKEN *token)
{
	if(token->token_type == TOKEN_TYPE_NL
			|| token->token_type == TOKEN_TYPE_EOF) {
		return 1;
	}
	return 0;
}

static int check_token_keyword(TOKEN *token)
{
	if(token->token_type == TOKEN_TYPE_PRINT ||
			token->token_type == TOKEN_TYPE_IF ||
			token->token_type == TOKEN_TYPE_GOTO ||
			token->token_type == TOKEN_TYPE_LET ||
			token->token_type == TOKEN_TYPE_INPUT ||
			token->token_type == TOKEN_TYPE_GOSUB ||
			token->token_type == TOKEN_TYPE_RETURN ||
			token->token_type == TOKEN_TYPE_CLEAR ||
			token->token_type == TOKEN_TYPE_END ||
			token->token_type == TOKEN_TYPE_REM ||
			token->token_type == TOKEN_TYPE_FOR ||
			token->token_type == TOKEN_TYPE_NEXT) {
		return 1;
	}
	return 0;
}
/* 文字列テーブルへの文字列追加 */
static BOOL add_string(VM_DATA *vm_data, LEX_DATA *lex_data, char *str)
{
	CZSTRING src, dest;

	CzString_init(&src);
	CzString_init(&dest);
	CzString_add_datas(&src, str, strlen(str) + 1);
#if defined(HAVE_ICONV) || defined(WIN32)
	if(!CzEncoding_convert(&vm_data->encMultbytetoUTF, &dest, &src)) {
		lex_data->err_code = ERR_OUTOFMEMORY;
		CzString_delete(&src);
		CzString_delete(&dest);
		return FALSE;
	}
#else
	for(i = 0; i < src.size; i++) {
		utf = src.data[i] & 0xff;
		if(!CzString_add_char(&dest, utf)) {
			lex_data->err_code = ERR_OUTOFMEMORY;
			CzString_delete(&src);
			CzString_delete(&dest);
			return FALSE;
		}
	}
#endif

	if(!CzMemory_add_datas(&vm_data->string_pool, (char*)&dest, sizeof(dest))) {
		lex_data->err_code = ERR_OUTOFMEMORY;
		CzString_delete(&src);
		CzString_delete(&dest);
		return FALSE;
	}
	CzString_delete(&src);
	st_string_pt++;

	return TRUE;
}

/* 命令ごとの解析ルーチンを呼ぶ */
static PARSE_RESULT parser_statement_call(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	int func_no;
	TOKEN *token;

	token = lex_get_token_skip_space(lex_data);

	func_no = token->token_type - TOKEN_TYPE_PRINT;
	if(func_no < 0) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}

	if(st_statement_func[func_no]) {
		lex_unget_token(lex_data, token);
		return st_statement_func[func_no](vm_data, lex_data);
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* 行番号テーブルへ追加 */
static PARSE_RESULT add_line_num_table(VM_DATA *vm_data, LEX_DATA *lex_data, int address)
{
	if(vm_data->line_num_table_last >= LINE_NUM_MAX) {
		lex_data->err_code = ERR_LINENUMMEM_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	vm_data->line_num_table[vm_data->line_num_table_last] = address;
	vm_data->line_num_table_last++;

	return PARSE_RESULT_TRUE;
}

/* ラベルの追加 */
static UINT32 add_label(VM_DATA *vm_data, char *label)
{
	size_t len;
	LABEL_TABLE label_table;
	UINT32 label_index;

	len = strlen(label) + 1;

	/* すでに登録済かどうか検索 */
	label_index = search_label(vm_data, label);
	if(label_index != (UINT32)-1) { /* 登録済なら */
		return label_index;
	}

	/* 追加登録 */
	label_index = vm_data->label_table.size / sizeof(LABEL_TABLE);
	Label_table_init(&label_table);
	if(!CzString_add_datas(&label_table.label, label, len)) {
		Label_table_delete(&label_table);
		return (UINT32)-1;
	}
	if(!CzMemory_add_datas(&vm_data->label_table, (char*)&label_table, sizeof(label_table))) {
		Label_table_delete(&label_table);
		return (UINT32)-1;
	}
	return label_index;
}

/* ラベルに割り当てられているアドレスを変更 */
static void change_label_address(VM_DATA *vm_data, UINT32 label_index, UINT32 address)
{
	LABEL_TABLE *item;

	item = (LABEL_TABLE*)vm_data->label_table.data;
	item[label_index].address = address;
}

/* tokenを符号無し整数に変換 */
static PARSE_RESULT conv_unsigned_int(TOKEN *token, int *retval)
{
	int pt = 0;

	*retval = 0;
	if(token->token[0] == '0' && token->token[1] != 0) {
		return PARSE_RESULT_FALSE;
	}
	while(token->token[pt] != 0) {
		if(!isdigit((int)token->token[pt])) {
			return PARSE_RESULT_FALSE;
		}
		*retval = *retval * 10 + token->token[pt] - '0';
		if(*retval & ~0x7fff) {
			return PARSE_RESULT_OVERFLOW;
		}
		pt++;
	}
	return PARSE_RESULT_TRUE;
}

/* プログラムメモリへコード追加 */
static BOOL add_progmem(VM_DATA *vm_data, LEX_DATA *lex_data, int data)
{
	vm_data->mem[vm_data->pc] = data;
	vm_data->pc++;
	vm_data->mem_last = vm_data->pc;
	if(vm_data->pc >= VMMEM_MAX) {
		lex_data->err_code = ERR_PROGMEM_OVERFLOW;
		return FALSE;
	}
	return TRUE;
}

/* 記号からオペコードへの変換 */
static enum opecode search_mark_to_opecode(TOKEN_TYPE mark)
{
	int i;
	i = 0;
	while(st_mark_list[i].mark != TOKEN_TYPE_NONE) {
		if(st_mark_list[i].mark == mark) {
			break;
		}
		i++;
	}
	return st_mark_list[i].opecode;
}

/* ラベルに割り当てられているアドレスを取得 */
UINT32 get_label_address(VM_DATA *vm_data, UINT32 label_index)
{
	LABEL_TABLE *item;

	item = (LABEL_TABLE*)vm_data->label_table.data;
	return item[label_index].address;
}

/* ラベルの検索 */
UINT32 search_label(VM_DATA *vm_data, char *label)
{
	size_t i;
	LABEL_TABLE *item;

	/* ラベルを検索 */
	item = (LABEL_TABLE*)vm_data->label_table.data;
	for(i = 0; i < vm_data->label_table.size / sizeof(LABEL_TABLE); i++) {
		if(Strcmp(item[i].label.data, label) == 0) {
			return i;
		}
	}

	return (UINT32)-1;
}


PARSE_RESULT parser(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	st_string_pt = 0;
	/* 改行とタブを文字列テーブルに追加しておく */
	NL_STRING_PT = st_string_pt;
	add_string(vm_data, lex_data, "\n");
	TAB_STRING_PT = st_string_pt;
	add_string(vm_data, lex_data, "\t");
	NEXT_ERROR_STRING_PT = st_string_pt;
	add_string(vm_data, lex_data, "NEXT without FOR in ");

	return parser_program(vm_data, lex_data);
}

/* <プログラム> -->  ε | <行> <プログラム> */
static PARSE_RESULT parser_program(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	UINT32 *label_item, addr;
	size_t i;

	lex_data->err_line_num = 1;
	/* トークン取得 */
	token = lex_get_token_skip_space(lex_data);
	while(token->token_type != TOKEN_TYPE_EOF) {
		lex_unget_token(lex_data, token);
		/* 行番号を追加 */
		if(add_line_num_table(vm_data, lex_data, vm_data->pc) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		res = parser_line(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		lex_data->err_line_num++;
		/* トークン取得 */
		token = lex_get_token_skip_space(lex_data);
	}

	if(!add_progmem(vm_data, lex_data, OP_STMT_END)) {
		return PARSE_RESULT_FALSE;
	}
	DPRINT("parser end");
	/* ラベルの実際のアドレスを設定 */
	label_item = (UINT32*)lex_data->label_address.data;
	for(i = 0; i < lex_data->label_address.size / sizeof(UINT32); i++) {
		/* ラベルのアドレスを取得 */
		addr = get_label_address(vm_data, vm_data->mem[label_item[i]]);
		if(addr == (UINT32)-1) {
			lex_data->err_code = ERR_UNDEFINEDLABEL;
			lex_data->err_line_num = search_editor_line_num(vm_data, label_item[i]);
			return PARSE_RESULT_FALSE;
		}
		/* 実際のアドレスを設定 */
		vm_data->mem[label_item[i]] = addr;
	}
	DPRINT("change addr");
	return PARSE_RESULT_TRUE;
}

/*
<行> -> <行番号> <行1> <行2> <CR>
<行> -> ‘#’ <全文字>* <CR> ? ? ? ? ? ? ? ? <- <CR>までスキップする処理をここで行う。
*/
static PARSE_RESULT parser_line(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_line");

	/* トークン取得 */
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_INT ||
			token->token_type == TOKEN_TYPE_MUL ||
			check_token_keyword(token) ||
			token->token_type == TOKEN_TYPE_WORD ||
			check_token_NL(token) ||
			token->token_type == TOKEN_TYPE_COLON) {
		lex_unget_token(lex_data, token);
		res = parser_linenum(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		res = parser_line1(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		res = parser_line2(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		token = lex_get_token_skip_space(lex_data);
		if(check_token_NL(token)) {
			return PARSE_RESULT_TRUE;
		} else {
			lex_data->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}
	} else if(token->token_type == TOKEN_TYPE_SHARP) {
		token = lex_get_token_skip_space(lex_data);
		while(!check_token_NL(token)) {
			token = lex_get_token_skip_space(lex_data);
		}
		return PARSE_RESULT_TRUE;
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
}


/* <行番号> -> ε | <符号無し整数> */
static PARSE_RESULT parser_linenum(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	int line_num;
	int label_index;

	DPRINT("parser_linenum");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_INT) {
		DPRINT("符号無し整数");
		if(conv_unsigned_int(token, &line_num) == PARSE_RESULT_TRUE) {
			label_index = add_label(vm_data, token->token);
			/* 行番号をラベルテーブルに追加 */
			if(label_index == (UINT32)-1) {
				lex_data->err_code = ERR_OUTOFMEMORY;
				return PARSE_RESULT_FALSE;
			}
			/* ラベルにアドレスを割り当てる */
			change_label_address(vm_data, label_index, vm_data->pc);
		} else {
			line_num = -1;
			/* トークンを戻す */
			lex_unget_token(lex_data, token);
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	
	return PARSE_RESULT_TRUE;
}

/* <行1> -> ε | <ラベル> <行3> */
static PARSE_RESULT parser_line1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int label_index;

	DPRINT("parser_line1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_MUL) {
		lex_unget_token(lex_data, token);
		res = parser_label(vm_data, lex_data, &label_index);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		res = parser_line3(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	
	return PARSE_RESULT_TRUE;
}

/* <行2> -> ε | <命令リスト> */
static PARSE_RESULT parser_line2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_line2");
	token = lex_get_token_skip_space(lex_data);
	if(check_token_keyword(token) ||
			token->token_type == TOKEN_TYPE_WORD ||
			token->token_type == TOKEN_TYPE_COLON) {
		lex_unget_token(lex_data, token);
		res = parser_statement_list(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}

	return PARSE_RESULT_TRUE;
}

/* <行3> -> ε | ‘:’ */
static PARSE_RESULT parser_line3(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;

	DPRINT("parser_line3");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_COLON) {
		DPRINT(":");
	} else {
		lex_unget_token(lex_data, token);
	}
	return PARSE_RESULT_TRUE;
}

/* <ラベル> -> ‘*’ <WORD> */
static PARSE_RESULT parser_label(VM_DATA *vm_data, LEX_DATA *lex_data, int *ret_index)
{
	TOKEN *token;
	int label_index;

	DPRINT("parser_label");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_MUL) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	token = lex_get_token(lex_data);
	if(token->token_type != TOKEN_TYPE_WORD) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	DPRINT("WORD");

	/* ラベルを追加 */
	label_index = add_label(vm_data, token->token);
	if(label_index == (UINT32)-1) {
		lex_data->err_code = ERR_OUTOFMEMORY;
		return PARSE_RESULT_FALSE;
	}

	/* ラベルにアドレスを設定 */
	change_label_address(vm_data, label_index, vm_data->pc);

	*ret_index = label_index;

	return PARSE_RESULT_TRUE;
}

/* <命令リスト> -> <命令リスト2> <命令リスト1> */
static PARSE_RESULT parser_statement_list(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	PARSE_RESULT res;

	DPRINT("parser_statement_list");
	res = parser_statement_list2(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	res = parser_statement_list1(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	
	return PARSE_RESULT_TRUE;
}

static PARSE_RESULT parser_statement(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int var_index;

	DPRINT("parser_statement");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_WORD) {
		lex_unget_token(lex_data, token);
		res = parser_statement_assignment(vm_data, lex_data, &var_index);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else if(!check_token_keyword(token)) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	} else {
		/* 予約語解析して実行 */
		lex_unget_token(lex_data, token);
		res = parser_statement_call(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {	
			return res;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* <命令リスト1> -> ε |  ‘:’ <命令リスト2> <命令リスト1> */
static PARSE_RESULT parser_statement_list1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_statement_list1");
	while(1) {
		token = lex_get_token(lex_data);
		if(token->token_type == TOKEN_TYPE_COLON) {
			res = parser_statement_list2(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
		} else {
			lex_unget_token(lex_data, token);
			break;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* <命令リスト2> -> ε | <命令> */
static PARSE_RESULT parser_statement_list2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_statement_list2");
	token = lex_get_token_skip_space(lex_data);
	if(check_token_keyword(token) ||
			token->token_type == TOKEN_TYPE_WORD) {
		lex_unget_token(lex_data, token);
		res = parser_statement(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	return PARSE_RESULT_TRUE;
}

/* <代入文> -> <変数> ‘=' <式> */
static PARSE_RESULT parser_statement_assignment(VM_DATA *vm_data, LEX_DATA *lex_data, int *var_index)
{
	TOKEN *token;
	PARSE_RESULT res;

	*var_index = -1;
	DPRINT("parser_statement_assignment");
	res = parser_var(vm_data, lex_data, var_index);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_EQ) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_expression(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	if(!(add_progmem(vm_data, lex_data, OP_POP_VAR) &&
			add_progmem(vm_data, lex_data, *var_index))) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <PRINT文> -> ‘PRINT' <式リスト> */
static PARSE_RESULT parser_print(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int add_nl;

	DPRINT("parser_print");
	add_nl = 1;
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_PRINT) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_expression_list(vm_data, lex_data, &add_nl);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}

	if(add_nl) {
		/* 出力先PUSH, 文字列出力OP(改行) */
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
					add_progmem(vm_data, lex_data, 1) &&
					add_progmem(vm_data, lex_data, OP_STMT_PRN_STRING) &&
				add_progmem(vm_data, lex_data, NL_STRING_PT))) {
			return PARSE_RESULT_FALSE;
		}
	}

	return PARSE_RESULT_TRUE;
}

/* <IF文> -> ‘IF' <式> <IF文3> <IF文1> */
static PARSE_RESULT parser_if(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int jmp_ncond_addr, jmp_addr;

	DPRINT("parser_if");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_IF) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_expression(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	/* 比較の結果がスタックにあるはずなので
	　 結果が偽のときのJMP OPを追加 */
	if(!add_progmem(vm_data, lex_data, OP_NCOND_JMP)) {
		return PARSE_RESULT_FALSE;
	}
	jmp_ncond_addr = vm_data->pc;
	/* 飛び先のアドレスを仮設定 */
	if(!add_progmem(vm_data, lex_data, 0)) {
		return PARSE_RESULT_FALSE;
	}
	
	res = parser_if3(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}

	/* IF文の次の行へのジャンプを追加 */
	/* JMPのオペコードを追加 */
	if(!add_progmem(vm_data, lex_data, OP_JMP)) {
		return PARSE_RESULT_FALSE;
	}
	jmp_addr = vm_data->pc;
	/* 飛び先のアドレスを仮設定 */
	if(!add_progmem(vm_data, lex_data, 0)) {
		return PARSE_RESULT_FALSE;
	}

	/* 偽のときの飛び先を設定 */
	vm_data->mem[jmp_ncond_addr] = vm_data->pc;

	res = parser_if1(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}

	/* 次の行へのジャンプ先を再設定 */
	vm_data->mem[jmp_addr] = vm_data->pc;

	return PARSE_RESULT_TRUE;
}

/* <GOTO文> -> 'GOTO’ <飛び先> */
static PARSE_RESULT parser_goto(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int label_index;

	DPRINT("parser_goto");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_GOTO) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_jump(vm_data, lex_data, &label_index);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	if(label_index == -1) {
		if(!add_progmem(vm_data, lex_data, OP_JMP_NUM)) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		if(!add_progmem(vm_data, lex_data, OP_JMP)) {
			return PARSE_RESULT_FALSE;
		}
		/* ラベルから実際のアドレスへ変換するアドレスを追加 */
		if(!CzMemory_add_datas(&lex_data->label_address, (char*)&vm_data->pc, sizeof(vm_data->pc))) {
			lex_data->err_code = ERR_OUTOFMEMORY;
			return PARSE_RESULT_FALSE;
		}
		/* アドレスをラベルのインデックスで仮設定 */
		if(!add_progmem(vm_data, lex_data, label_index)) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* <LET文> -> 'LET' <変数> ‘=' <式> */
static PARSE_RESULT parser_let(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	int var_index;

	DPRINT("parser_let");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_LET) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return parser_statement_assignment(vm_data, lex_data, &var_index);
}

/* <INPUT文> -> 'INPUT' <変数リスト> */
static PARSE_RESULT parser_input(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int count;

	DPRINT("parser_input");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_INPUT) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_var_list(vm_data, lex_data, &count);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}

	if(!(add_progmem(vm_data, lex_data, OP_INPUT) &&
			add_progmem(vm_data, lex_data, count))) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <GOSUB文> -> ‘GOSUB' <飛び先> */
static PARSE_RESULT parser_gosub(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int label_index;

	DPRINT("parser_gosub");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_GOSUB) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_jump(vm_data, lex_data, &label_index);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	if(label_index == -1) {
		if(!add_progmem(vm_data, lex_data, OP_CALL_NUM)) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		if(!add_progmem(vm_data, lex_data, OP_CALL)) {
			return PARSE_RESULT_FALSE;
		}
		/* ラベルから実際のアドレスへ変換するアドレスを追加 */
		if(!CzMemory_add_datas(&lex_data->label_address, (char*)&vm_data->pc, sizeof(vm_data->pc))) {
			lex_data->err_code = ERR_OUTOFMEMORY;
			return PARSE_RESULT_FALSE;
		}
		/* アドレスをラベルのインデックスで仮設定 */
		if(!add_progmem(vm_data, lex_data, label_index)) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* <RETURN文> -> 'RETURN' */
static PARSE_RESULT parser_return(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	
	DPRINT("parser_return");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_RETURN) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	if(!add_progmem(vm_data, lex_data, OP_RET)) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}
/* <CLEAR文> -> ‘CLEAR' */
static PARSE_RESULT parser_clear(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	
	DPRINT("parser_clear");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_CLEAR) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	if(!add_progmem(vm_data, lex_data, OP_STMT_CLEAR)) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <END文> -> 'END' */
static PARSE_RESULT parser_end(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	
	DPRINT("parser_end");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_END) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	if(!add_progmem(vm_data, lex_data, OP_STMT_END)) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}
static PARSE_RESULT parser_rem(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;

	DPRINT("parser_rem");
	do {
		token = lex_get_token_skip_space(lex_data);
	} while(!check_token_NL(token));
	lex_unget_token(lex_data, token);
	return PARSE_RESULT_TRUE;
}

/* <FOR文> -> ‘FOR’ <代入文> ‘TO’ <式> <FOR文1> */
static PARSE_RESULT parser_for(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int var_index;

	DPRINT("parser_for");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_FOR) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_statement_assignment(vm_data, lex_data, &var_index);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	/* 初期値をスタックに積む */
	if(!(add_progmem(vm_data, lex_data, OP_PUSH_VAR)
			&& add_progmem(vm_data, lex_data, var_index))) {
		return PARSE_RESULT_FALSE;
	}

	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_TO) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parser_expression(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	res = parser_for1(vm_data, lex_data, var_index);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	return PARSE_RESULT_TRUE;
}

/* <FOR文1> -> ε | ‘STEP’ <式> */
static PARSE_RESULT parser_for1(VM_DATA *vm_data, LEX_DATA *lex_data, int var_index)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_for1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_STEP) {
		res = parser_expression(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) 
				&& add_progmem(vm_data, lex_data, 1))) {
			return PARSE_RESULT_FALSE;
		}
	}
	/* 変数インデックスをプッシュ */
	if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) 
			&& add_progmem(vm_data, lex_data, var_index))) {
		return PARSE_RESULT_FALSE;
	}
	/* 戻り先アドレスをプッシュ */
	if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
		add_progmem(vm_data, lex_data, vm_data->pc + 1))) {
			return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <NEXT文> -> ‘NEXT’ <NEXT文1> */
static PARSE_RESULT parser_next(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int var_index;
	int count;
	int break_flag;
	UINT32 tmp_pc;

	DPRINT("parser_next");
	count = 0;
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_NEXT) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	while(1) {
		break_flag = 0;
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type != TOKEN_TYPE_WORD) {
			break_flag = 1;
		}
		lex_unget_token(lex_data, token);
		if(break_flag && count != 0) {
			break;
		}

		if(!break_flag) {
			res = parser_var(vm_data, lex_data, &var_index);
			if(res != PARSE_RESULT_TRUE) {
				lex_data->err_code = ERR_SYNTAX;
				return PARSE_RESULT_FALSE;
			}
			count++;

			if(!(add_progmem(vm_data, lex_data, OP_REPUSH) &&
					add_progmem(vm_data, lex_data, -2) &&
					add_progmem(vm_data, lex_data, 1))) {
				return PARSE_RESULT_FALSE;
			}
			/* 変数インデックスをスタックするコードを追加 */
			if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, var_index))) {
					return PARSE_RESULT_FALSE;
			}
			/* 変数がFORで指定された変数と一致するかチェックするコードを追加 */
			if(!add_progmem(vm_data, lex_data, OP_CMP_EQ)) {
				return PARSE_RESULT_FALSE;
			}
			if(!add_progmem(vm_data, lex_data, OP_COND_JMP)) {
				return PARSE_RESULT_FALSE;
			}
			tmp_pc = vm_data->pc;
			if(!add_progmem(vm_data, lex_data, 0)) {
				return PARSE_RESULT_FALSE;
			}
			/* エラーメッセージを出力するコードを追加 */
			if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, 2) &&
				add_progmem(vm_data, lex_data, OP_STMT_PRN_STRING) &&
				add_progmem(vm_data, lex_data, NEXT_ERROR_STRING_PT) &&
				add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, 2) &&
				add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, (UINT32)lex_data->err_line_num) &&
				add_progmem(vm_data, lex_data, OP_STMT_PRN) &&
				add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, 2) &&
				add_progmem(vm_data, lex_data, OP_STMT_PRN_STRING) &&
				add_progmem(vm_data, lex_data, NL_STRING_PT) &&
				add_progmem(vm_data, lex_data, OP_STMT_END))) {
					return PARSE_RESULT_FALSE;
			}
			vm_data->mem[tmp_pc] = vm_data->pc;
		}
		/* STEPの値加算する */
		if(!(add_progmem(vm_data, lex_data, OP_REPUSH) &&
			add_progmem(vm_data, lex_data, (UINT32)-2) &&
			add_progmem(vm_data, lex_data, 1) &&
			add_progmem(vm_data, lex_data, OP_PUSH_VAR2) &&
			add_progmem(vm_data, lex_data, OP_REPUSH) &&
			add_progmem(vm_data, lex_data, (UINT32)-4) &&
			add_progmem(vm_data, lex_data, 1) &&
			add_progmem(vm_data, lex_data, OP_ADD) &&
			add_progmem(vm_data, lex_data, OP_REPUSH) &&
			add_progmem(vm_data, lex_data, (UINT32)-3) &&
			add_progmem(vm_data, lex_data, 1) &&
			add_progmem(vm_data, lex_data, OP_POP_VAR2))) {
				return PARSE_RESULT_FALSE;
		}

		/* 範囲チェック */
		if(!(add_progmem(vm_data, lex_data, OP_REPUSH) &&
			add_progmem(vm_data, lex_data, (UINT32)-5) &&
			add_progmem(vm_data, lex_data, 2) &&
			add_progmem(vm_data, lex_data, OP_REPUSH) &&
			add_progmem(vm_data, lex_data, (UINT32)-4) &&
			add_progmem(vm_data, lex_data, 1) &&
			add_progmem(vm_data, lex_data, OP_PUSH_VAR2) &&
			add_progmem(vm_data, lex_data, OP_CMP_RANGE))) {
				return PARSE_RESULT_FALSE;
		}
		if(!add_progmem(vm_data, lex_data, OP_NCOND_JMP)) {
			return PARSE_RESULT_FALSE;
		}
		tmp_pc = vm_data->pc;
		if(!add_progmem(vm_data, lex_data, 0)) {
			return PARSE_RESULT_FALSE;
		}
		if(!(add_progmem(vm_data, lex_data, OP_REPUSH) &&
			add_progmem(vm_data, lex_data, (UINT32)-1) &&
			add_progmem(vm_data, lex_data, 1) &&
			add_progmem(vm_data, lex_data, OP_POP_PC))) {
				return PARSE_RESULT_FALSE;
		}
		vm_data->mem[tmp_pc] = vm_data->pc;

		if(!(add_progmem(vm_data, lex_data, OP_POP_DMY) &&
			add_progmem(vm_data, lex_data, OP_POP_DMY) &&
			add_progmem(vm_data, lex_data, OP_POP_DMY) &&
			add_progmem(vm_data, lex_data, OP_POP_DMY) &&
			add_progmem(vm_data, lex_data, OP_POP_DMY))) {
			return PARSE_RESULT_FALSE;
		}
		if(break_flag && count == 0) {
			break;
		}

		token = lex_get_token_skip_space(lex_data);
		if(token->token_type != TOKEN_TYPE_COMMA) {
			lex_unget_token(lex_data, token);
			break;
		}
	}

	return PARSE_RESULT_TRUE;
}

#if 0
/* <NEXT文1> -> ε | <変数リスト> */
static PARSE_RESULT parser_next1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int count;

	DPRINT("parser_next1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_WORD) {
		lex_unget_token(lex_data, token);
		res = parser_var_list(vm_data, lex_data, &count);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	return PARSE_RESULT_TRUE;
}
#endif

/* <変数> -> <WORD>                                           <-- <WORD>は1文字かつアルファベットである必要がある。 */
static PARSE_RESULT parser_var(VM_DATA *vm_data, LEX_DATA *lex_data, int *var_index)
{
	TOKEN *token;

	*var_index = -1;
	DPRINT("parser_var");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_WORD) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	if(strlen(token->token) != 1) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	DPRINT(token->token);

	if(token->token[0] >= 'a' && token->token[0] <= 'z') {
		*var_index = token->token[0] - 'a';
		return PARSE_RESULT_TRUE;
	}
	if(token->token[0] >= 'A' && token->token[0] <= 'Z') {
		*var_index = token->token[0] - 'A';
		return PARSE_RESULT_TRUE;
	}
	return PARSE_RESULT_FALSE;
}

/* <式> -> <項1>  <式2> */
static PARSE_RESULT parser_expression(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	PARSE_RESULT res;

	DPRINT("parser_expression");
	res = parser_item1(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	res = parser_expression2(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	return PARSE_RESULT_TRUE;
}

/* <項1> -> <項2> <項1_2> */
static PARSE_RESULT parser_item1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	PARSE_RESULT res;

	DPRINT("parser_item1");
	res = parser_item2(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	res = parser_item1_2(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	return PARSE_RESULT_TRUE;
}

/* <式2> -> ε | <式1> <項1> <式2> */
static PARSE_RESULT parser_expression2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	TOKEN_TYPE backup_type;
	enum opecode ope;

	DPRINT("parser_expression2");
	while(1) {
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_EQ ||
				token->token_type == TOKEN_TYPE_LT ||
				token->token_type == TOKEN_TYPE_GT ||
				token->token_type == TOKEN_TYPE_LE ||
				token->token_type == TOKEN_TYPE_GE ||
				token->token_type == TOKEN_TYPE_NE) {
			backup_type = token->token_type;
			lex_unget_token(lex_data, token);
			res = parser_expression1(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
			res = parser_item1(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}

			/* オペコードを追加 */
			ope = search_mark_to_opecode(backup_type);
			if(ope == OP_NOP) {
				lex_data->err_code = ERR_SYNTAX;
				return PARSE_RESULT_FALSE;
			}
			if(!add_progmem(vm_data, lex_data, ope)) {
				return PARSE_RESULT_FALSE;
			}
		} else {
			lex_unget_token(lex_data, token);
			break;
		}
	}
			
	return PARSE_RESULT_TRUE;
}

/* <項2> -> <基本式> <項2_2> */
static PARSE_RESULT parser_item2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	PARSE_RESULT res;

	DPRINT("parser_item2");
	res = parser_basis_expression(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	res = parser_item2_2(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	return PARSE_RESULT_TRUE;
}

/* <項1_2> -> ε | <項1_1> <項2> <項1_2> */
static PARSE_RESULT parser_item1_2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	TOKEN_TYPE backup_type;
	enum opecode ope;

	DPRINT("parser_item1_2");
	while(1) {
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_PLUS ||
				token->token_type == TOKEN_TYPE_MINUS) {
			backup_type = token->token_type;
			lex_unget_token(lex_data, token);
			res = parser_item1_1(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
			res = parser_item2(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}

			/* オペコードを追加 */
			ope = search_mark_to_opecode(backup_type);
			if(ope == OP_NOP) {
				lex_data->err_code = ERR_SYNTAX;
				return PARSE_RESULT_FALSE;
			}
			if(!add_progmem(vm_data, lex_data, ope)) {
				return PARSE_RESULT_FALSE;
			}
		} else {
			lex_unget_token(lex_data, token);
			break;
		}
	}
			
	return PARSE_RESULT_TRUE;
}

/* <基本式> -> <整数> | "(" <式> ")" | <変数> */
static PARSE_RESULT parser_basis_expression(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int var_index;

	DPRINT("parser_basis_expression");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_LPARENTHESES) {
		res = parser_expression(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type != TOKEN_TYPE_RPARENTHESES) {
			lex_data->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}
	} else if(token->token_type == TOKEN_TYPE_WORD) {
		lex_unget_token(lex_data, token);
		res = parser_var(vm_data, lex_data, &var_index);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_VAR) &&
				add_progmem(vm_data, lex_data, var_index))) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		lex_unget_token(lex_data, token);
		res = parser_int(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* <項2_2> -> ε | <項2_1> <基本式> <項2_2> */
static PARSE_RESULT parser_item2_2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	TOKEN_TYPE backup_type;
	enum opecode ope;

	DPRINT("parser_item2_2");
	while(1) {
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_MUL ||
				token->token_type == TOKEN_TYPE_DIV) {
			backup_type = token->token_type;
			lex_unget_token(lex_data, token);
			res = parser_item2_1(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
			res = parser_basis_expression(vm_data, lex_data);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}

			/* オペコードを追加 */
			ope = search_mark_to_opecode(backup_type);
			if(ope == OP_NOP) {
				lex_data->err_code = ERR_SYNTAX;
				return PARSE_RESULT_FALSE;
			}
			if(!add_progmem(vm_data, lex_data, ope)) {
				return PARSE_RESULT_FALSE;
			}
		} else {
			lex_unget_token(lex_data, token);
			break;
		}
	}
			
	return PARSE_RESULT_TRUE;
}

/* <整数> -> <整数1> <符号無し整数> */
static PARSE_RESULT parser_int(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int fugo, val;

	DPRINT("parser_int");
#if 1
	fugo = 1;
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_MINUS) {
		fugo = -1;
	} else if(token->token_type == TOKEN_TYPE_PLUS) {
		fugo = 1;
	} else {
		lex_unget_token(lex_data, token);
	}
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_INT) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = conv_unsigned_int(token, &val);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
			add_progmem(vm_data, lex_data, val * fugo))) {
		return PARSE_RESULT_FALSE;
	}
#else
	res = parser_int1(vm_data, lex_data);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type != TOKEN_TYPE_INT) {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	DPRINT("符号無し整数");
#endif
	return PARSE_RESULT_TRUE;
}


#if 0 /********/
/* <整数1> -> ε | <符号> */
static PARSE_RESULT parser_int1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_int1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_PLUS ||
			token->token_type == TOKEN_TYPE_MINUS) {
		lex_unget_token(lex_data, token);
		res = parser_code(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	return PARSE_RESULT_TRUE;
}
#endif

#if 0 /**********/
/* <符号> -> ‘+’ | ‘-‘ */
static PARSE_RESULT parser_code(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_code");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_PLUS) {
		DPRINT("+");
	} else if(token->token_type == TOKEN_TYPE_MINUS) {
		DPRINT("-");
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}
#endif

static PARSE_RESULT parser_item2_1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;

	DPRINT("parser_item2_1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_MUL) {
		DPRINT("*");
	} else if(token->token_type == TOKEN_TYPE_DIV) {
		DPRINT("/");
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <項1_1> -> “+” | “-" */
static PARSE_RESULT parser_item1_1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;

	DPRINT("parser_item1_1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_PLUS) {
		DPRINT("+");
	} else if(token->token_type == TOKEN_TYPE_MINUS) {
		DPRINT("-");
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <式1> -> “=“ | ”<“ | ”>” | "<=“ | ">=“ | "<>” */
static PARSE_RESULT parser_expression1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;

	DPRINT("parser_expression1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_EQ) {
		DPRINT("=");
	} else if(token->token_type == TOKEN_TYPE_LT) {
		DPRINT("<");
	} else if(token->token_type == TOKEN_TYPE_GT) {
		DPRINT(">");
	} else if(token->token_type == TOKEN_TYPE_LE) {
		DPRINT("<=");
	} else if(token->token_type == TOKEN_TYPE_GE) {
		DPRINT(">=");
	} else if(token->token_type == TOKEN_TYPE_NE) {
		DPRINT("<>");
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <IF文1> -> ε | ‘ELSE’ <IF文2> */
static PARSE_RESULT parser_if1(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_if1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_ELSE) {
		res = parser_if2(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	return PARSE_RESULT_TRUE;
}

/* <IF文2> -> <ラベル> | <符号無し整数> | <命令リスト> */
static PARSE_RESULT parser_if2(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;
	int line_num;
	int label_index;

	DPRINT("parser_if2");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_INT) {
		DPRINT("符号無し整数");
		DPRINT(token->token);
		/* 行番号をPUSH */
		res = conv_unsigned_int(token, &line_num);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, line_num))) {
			return PARSE_RESULT_FALSE;
		}
		/* JMP OPを追加 */
		if(!add_progmem(vm_data, lex_data, OP_JMP_NUM)) {
			return PARSE_RESULT_FALSE;
		}
	} else if(token->token_type == TOKEN_TYPE_MUL) {
		lex_unget_token(lex_data, token);
		res = parser_label(vm_data, lex_data, &label_index);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		if(!add_progmem(vm_data, lex_data, OP_JMP)) {
			return PARSE_RESULT_FALSE;
		}
		/* ラベルから実際のアドレスへ変換するアドレスを追加 */
		if(!CzMemory_add_datas(&lex_data->label_address, (char*)&vm_data->pc, sizeof(vm_data->pc))) {
			lex_data->err_code = ERR_OUTOFMEMORY;
			return PARSE_RESULT_FALSE;
		}
		/* アドレスをラベルのインデックスで仮設定 */
		if(!add_progmem(vm_data, lex_data, label_index)) {
			return PARSE_RESULT_FALSE;
		}
		
	} else {
		lex_unget_token(lex_data, token);
		res = parser_statement_list(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	}
	return PARSE_RESULT_TRUE;
}
/* <IF文3> -> ‘THEN' <IF文2> | ‘GOTO’ <飛び先> */
static PARSE_RESULT parser_if3(VM_DATA *vm_data, LEX_DATA *lex_data)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_if3");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_THEN) {
		res = parser_if2(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else if(token->token_type == TOKEN_TYPE_GOTO) {
		lex_unget_token(lex_data, token);
		res = parser_goto(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}


/* <式リスト> -> ε | <式リスト1> ?<式リスト4> */
static PARSE_RESULT parser_expression_list(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_expression_list");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_STRING ||
			token->token_type == TOKEN_TYPE_PLUS ||
			token->token_type == TOKEN_TYPE_MINUS ||
			token->token_type == TOKEN_TYPE_LPARENTHESES ||
			token->token_type == TOKEN_TYPE_WORD) {
		lex_unget_token(lex_data, token);
		res = parser_expression_list1(vm_data, lex_data, add_nl);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		res = parser_expression_list4(vm_data, lex_data, add_nl);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}

	return PARSE_RESULT_TRUE;
}

static PARSE_RESULT parser_expression_list1(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_expression_list1");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_STRING) {
		DPRINT("文字列");
		/* 出力先PUSH, 文字列出力OP */
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, 1) &&
				add_progmem(vm_data, lex_data, OP_STMT_PRN_STRING) &&
				add_progmem(vm_data, lex_data, st_string_pt) &&
				add_string(vm_data, lex_data, token->token))) {
			return PARSE_RESULT_FALSE;
		}
		*add_nl = 1;
	} else if(token->token_type == TOKEN_TYPE_PLUS ||
				token->token_type == TOKEN_TYPE_MINUS ||
				token->token_type == TOKEN_TYPE_LPARENTHESES ||
				token->token_type == TOKEN_TYPE_WORD ||
				token->token_type == TOKEN_TYPE_INT) {
		/* 出力先PUSH */
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, 1))) {
			return PARSE_RESULT_FALSE;
		}
		lex_unget_token(lex_data, token);
		res = parser_expression(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
		/* 数値出力OP */
		if(!add_progmem(vm_data, lex_data, OP_STMT_PRN)) {
			return PARSE_RESULT_FALSE;
		}
		*add_nl = 1;
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <式リスト2> -> “,” | “;” */
static PARSE_RESULT parser_expression_list2(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl)
{
	TOKEN *token;

	DPRINT("parser_expression_list2");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_COMMA) {
		DPRINT(",");
		/* 出力先PUSH, 文字列出力OP */
		if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
				add_progmem(vm_data, lex_data, 1) &&
				add_progmem(vm_data, lex_data, OP_STMT_PRN_STRING) &&
				add_progmem(vm_data, lex_data, TAB_STRING_PT))) {
			return PARSE_RESULT_FALSE;
		}
		*add_nl = 1;
	} else if(token->token_type == TOKEN_TYPE_SEMICOLON) {
		DPRINT(";");
		*add_nl = 0;
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* <式リスト3> -> ε  | <式リスト1> */
static PARSE_RESULT parser_expression_list3(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_expression_list3");
	token = lex_get_token_skip_space(lex_data);
	if(token->token_type == TOKEN_TYPE_PLUS ||
			token->token_type == TOKEN_TYPE_MINUS ||
			token->token_type == TOKEN_TYPE_LPARENTHESES ||
			token->token_type == TOKEN_TYPE_WORD ||
			token->token_type == TOKEN_TYPE_STRING ||
			token->token_type == TOKEN_TYPE_INT) {
		lex_unget_token(lex_data, token);
		res = parser_expression_list1(vm_data, lex_data, add_nl);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_unget_token(lex_data, token);
	}
	return PARSE_RESULT_TRUE;
}

/* <式リスト4> -> ε  | <式リスト2> <式リスト3> <式リスト4> */
static PARSE_RESULT parser_expression_list4(VM_DATA *vm_data, LEX_DATA *lex_data, int *add_nl)
{
	TOKEN *token;
	PARSE_RESULT res;

	DPRINT("parser_expression_list4");
	while(1) {
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_COMMA ||
				token->token_type == TOKEN_TYPE_SEMICOLON) {
			lex_unget_token(lex_data, token);
			res = parser_expression_list2(vm_data, lex_data, add_nl);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
			res = parser_expression_list3(vm_data, lex_data, add_nl);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
		} else {
			lex_unget_token(lex_data, token);
			break;
		}
	}
		
	return PARSE_RESULT_TRUE;
}

/* <変数リスト> ->  <変数> <変数リスト1> */
static PARSE_RESULT parser_var_list(VM_DATA *vm_data, LEX_DATA *lex_data, int *ret_count)
{
	PARSE_RESULT res;
	int var_index;

	DPRINT("parser_var_list");
	*ret_count = 0;
	res = parser_var(vm_data, lex_data, &var_index);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
			add_progmem(vm_data, lex_data, var_index))) {
		return PARSE_RESULT_FALSE;
	}
	(*ret_count)++;
	res = parser_var_list1(vm_data, lex_data, ret_count);
	if(res != PARSE_RESULT_TRUE) {
		return res;
	}
	return PARSE_RESULT_TRUE;
}

/* <変数リスト1> -> ε | "," <変数> <変数リスト1> */
static PARSE_RESULT parser_var_list1(VM_DATA *vm_data, LEX_DATA *lex_data, int *ret_count)
{
	TOKEN *token;
	PARSE_RESULT res;
	int var_index;

	DPRINT("parser_var_list1");
	while(1) {
		token = lex_get_token_skip_space(lex_data);
		if(token->token_type == TOKEN_TYPE_COMMA) {
			res = parser_var(vm_data, lex_data, &var_index);
			if(res != PARSE_RESULT_TRUE) {
				return res;
			}
			if(!(add_progmem(vm_data, lex_data, OP_PUSH_INT) &&
					add_progmem(vm_data, lex_data, var_index))) {
				return PARSE_RESULT_FALSE;
			}
			(*ret_count)++;
		} else {
			lex_unget_token(lex_data, token);
			break;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* <飛び先> -> <式> | <ラベル> */
static PARSE_RESULT parser_jump(VM_DATA *vm_data, LEX_DATA *lex_data, int *label_index)
{
	TOKEN *token;
	PARSE_RESULT res;

	*label_index = -1;
	DPRINT("parser_jump");
	token = lex_get_token_skip_space(lex_data);
	if( token->token_type == TOKEN_TYPE_PLUS ||
			token->token_type == TOKEN_TYPE_MINUS ||
			token->token_type == TOKEN_TYPE_LPARENTHESES ||
			token->token_type == TOKEN_TYPE_WORD || 
			token->token_type == TOKEN_TYPE_INT) {
		lex_unget_token(lex_data, token);
		res = parser_expression(vm_data, lex_data);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else if(token->token_type == TOKEN_TYPE_MUL) {
		lex_unget_token(lex_data, token);
		res = parser_label(vm_data, lex_data, label_index);
		if(res != PARSE_RESULT_TRUE) {
			return res;
		}
	} else {
		lex_data->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}
